# -*- coding: utf-8 -*-
"""
This modules provides functions to calculate correlations between spike trains.

.. autosummary::
    :toctree: _toctree/spike_train_correlation

    covariance
    correlation_coefficient
    cross_correlation_histogram
    spike_time_tiling_coefficient
    spike_train_timescale

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
from __future__ import division, print_function, unicode_literals

import warnings

import neo
import numpy as np
import quantities as pq
import scipy.signal
from scipy import integrate
from elephant.utils import check_neo_consistency


__all__ = [
    "covariance",
    "correlation_coefficient",
    "cross_correlation_histogram",
    "spike_time_tiling_coefficient",
    "spike_train_timescale"
]

# The highest sparsity of the `BinnedSpikeTrain` matrix for which
# memory-efficient (sparse) implementation of `covariance()` is faster than
# with the corresponding numpy dense array.
_SPARSITY_MEMORY_EFFICIENT_THR = 0.1


class _CrossCorrHist(object):
    """
    Cross-correlation histogram for `BinnedSpikeTrain`s.
    This class is used inside :func:`cross_correlation_histogram` function
    and is not meant to be used outside of it.

    Parameters
    ----------
    binned_spiketrain_i, binned_spiketrain_j :
        elephant.conversion.BinnedSpikeTrain
        Binned spike trains to cross-correlate. The two spike trains must
        have the same `t_start` and `t_stop`.
    window : list or tuple
        List of integers - (left_edge, right_edge).
        Refer to the docs of `cross_correlation_histogram()`.
    """

    def __init__(self, binned_spiketrain_i, binned_spiketrain_j, window):
        self.binned_spiketrain_i = binned_spiketrain_i
        self.binned_spiketrain_j = binned_spiketrain_j
        self.window = window

    @staticmethod
    def get_valid_lags(binned_spiketrain_i, binned_spiketrain_j):
        """
        Computes the lags at which the cross-correlation
        of the input spiketrains can be calculated with full
        overlap.

        Parameters
        ----------
        binned_spiketrain_i, binned_spiketrain_j :
            elephant.conversion.BinnedSpikeTrain
            Binned spike trains to cross-correlate. The input spike trains can
            have any `t_start` and `t_stop`.

        Returns
        -------
        lags : np.ndarray
            Array of lags at which the cross-correlation can be computed
            at full overlap (valid mode).
        """

        bin_size = binned_spiketrain_i._bin_size

        # see cross_correlation_histogram for the examples
        if binned_spiketrain_i.n_bins < binned_spiketrain_j.n_bins:
            # ex. 1) lags range: [-2, 5] ms
            # ex. 2) lags range: [1, 2] ms
            left_edge = (binned_spiketrain_j._t_start -
                         binned_spiketrain_i._t_start) / bin_size
            right_edge = (binned_spiketrain_j._t_stop -
                          binned_spiketrain_i._t_stop) / bin_size
        else:
            # ex. 3) lags range: [-1, 3] ms
            left_edge = (binned_spiketrain_j._t_stop -
                         binned_spiketrain_i._t_stop) / bin_size
            right_edge = (binned_spiketrain_j._t_start -
                          binned_spiketrain_i._t_start) / bin_size
        right_edge = int(right_edge)
        left_edge = int(left_edge)
        lags = np.arange(left_edge, right_edge + 1, dtype=np.int32)

        return lags

    def correlate_memory(self, cch_mode):
        """
        Slow, but memory-safe mode.

        Return
        -------
        cross_corr : np.ndarray
            Cross-correlation of `self.binned_spiketrain1` and
            `self.binned_spiketrain2`.
        """
        st1_spmat = self.binned_spiketrain_i.sparse_matrix
        st2_spmat = self.binned_spiketrain_j.sparse_matrix
        left_edge, right_edge = self.window

        # extract the nonzero column indices of 1-d matrices
        st1_bin_idx_unique = st1_spmat.nonzero()[1]
        st2_bin_idx_unique = st2_spmat.nonzero()[1]

        # 'valid' mode requires bins correction due to the shift in t_starts
        # 'full' and 'pad' modes don't need this correction
        if cch_mode == "valid":
            if self.binned_spiketrain_i.n_bins > \
                    self.binned_spiketrain_j.n_bins:
                st2_bin_idx_unique += right_edge
            else:
                st2_bin_idx_unique += left_edge

        st1_spmat = st1_spmat.data
        st2_spmat = st2_spmat.data

        # Initialize the counts to an array of zeros,
        # and the bin IDs to integers
        # spanning the time axis
        nr_lags = right_edge - left_edge + 1
        cross_corr = np.zeros(nr_lags)

        # Compute the CCH at lags in left_edge,...,right_edge only
        for idx, i in enumerate(st1_bin_idx_unique):
            il = np.searchsorted(st2_bin_idx_unique, left_edge + i)
            ir = np.searchsorted(st2_bin_idx_unique,
                                 right_edge + i, side='right')
            timediff = st2_bin_idx_unique[il:ir] - i
            assert ((timediff >= left_edge) & (
                timediff <= right_edge)).all(), \
                'Not all the entries of cch lie in the window'
            cross_corr[timediff - left_edge] += (
                st1_spmat[idx] * st2_spmat[il:ir])
            st2_bin_idx_unique = st2_bin_idx_unique[il:]
            st2_spmat = st2_spmat[il:]
        return cross_corr

    def correlate_speed(self, cch_mode):
        """
        Fast, but might require a lot of memory.

        Parameters
        ----------
        cch_mode : str
            Cross-correlation mode.

        Returns
        -------
        cross_corr : np.ndarray
            Cross-correlation of `self.binned_spiketrain1` and
            `self.binned_spiketrain2`.
        """
        # Retrieve the array of the binned spike trains
        st1_arr = self.binned_spiketrain_i.to_array()[0]
        st2_arr = self.binned_spiketrain_j.to_array()[0]
        left_edge, right_edge = self.window
        if cch_mode == 'pad':
            # Zero padding to stay between left_edge and right_edge
            pad_width = min(max(-left_edge, 0), max(right_edge, 0))
            st2_arr = np.pad(st2_arr, pad_width=pad_width, mode='constant')
            cch_mode = 'valid'
        # Cross correlate the spike trains
        cross_corr = scipy.signal.fftconvolve(st2_arr, st1_arr[::-1],
                                              mode=cch_mode)
        # convolution of integers is integers
        cross_corr = np.round(cross_corr)
        return cross_corr

    def border_correction(self, cross_corr):
        """
        Parameters
        ----------
        cross_corr : np.ndarray
            Cross-correlation array. The output of `self.correlate_speed()`
            or `self.correlate_memory()`.

        Returns
        -------
        np.ndarray
            Cross-correlation array with the border correction applied.
        """
        min_num_bins = min(self.binned_spiketrain_i.n_bins,
                           self.binned_spiketrain_j.n_bins)
        left_edge, right_edge = self.window
        valid_lags = _CrossCorrHist.get_valid_lags(self.binned_spiketrain_i,
                                                   self.binned_spiketrain_j)
        lags_to_compute = np.arange(left_edge, right_edge + 1)
        outer_subtraction = np.subtract.outer(lags_to_compute, valid_lags)
        min_distance_from_window = np.abs(outer_subtraction).min(axis=1)
        n_values_fall_in_window = min_num_bins - min_distance_from_window
        correction = float(min_num_bins) / n_values_fall_in_window
        return cross_corr * correction

    def cross_correlation_coefficient(self, cross_corr):
        """
        Normalizes the CCH to obtain the cross-correlation coefficient
        function, ranging from -1 to 1.

        Parameters
        ----------
        cross_corr : np.ndarray
            Cross-correlation array. The output of `self.correlate_speed()`
            or `self.correlate_memory()`.

        Notes
        -----
        See Notes in `cross_correlation_histogram()`.

        Returns
        -------
        np.ndarray
            Normalized cross-correlation array in range `[-1, 1]`.
        """
        max_num_bins = max(self.binned_spiketrain_i.n_bins,
                           self.binned_spiketrain_j.n_bins)
        n_spikes1 = self.binned_spiketrain_i.get_num_of_spikes()
        n_spikes2 = self.binned_spiketrain_j.get_num_of_spikes()
        data1 = self.binned_spiketrain_i.sparse_matrix.data
        data2 = self.binned_spiketrain_j.sparse_matrix.data
        ii = data1.dot(data1)
        jj = data2.dot(data2)
        cov_mean = n_spikes1 * n_spikes2 / max_num_bins
        std_xy = np.sqrt((ii - n_spikes1 ** 2. / max_num_bins) * (
            jj - n_spikes2 ** 2. / max_num_bins))
        cross_corr_normalized = (cross_corr - cov_mean) / std_xy
        return cross_corr_normalized

    def kernel_smoothing(self, cross_corr_array, kernel):
        """
        Performs 1-d convolution with the `kernel`.

        Parameters
        ----------
        cross_corr_array : np.ndarray
            Cross-correlation array. The output of `self.correlate_speed()`
            or `self.correlate_memory()`.
        kernel : list
            1-d kernel.

        Returns
        -------
        np.ndarray
            Smoothed array.
        """
        left_edge, right_edge = self.window
        kern_len_max = abs(left_edge) + abs(right_edge) + 1
        # Define the kern for smoothing as an ndarray
        if len(kernel) > kern_len_max:
            raise ValueError(
                'The length of the kernel {} cannot be larger than the '
                'length {} of the resulting CCH.'.format(len(kernel),
                                                         kern_len_max))
        kernel = np.divide(kernel, kernel.sum())
        # Smooth the cross-correlation histogram with the kern
        return np.convolve(cross_corr_array, kernel, mode='same')


def covariance(binned_spiketrain, binary=False, fast=True):
    r"""
    Calculate the NxN matrix of pairwise covariances between all combinations
    of N binned spike trains.

    For each pair of spike trains :math:`(i,j)`, the covariance :math:`C[i,j]`
    is obtained by binning :math:`i` and :math:`j` at the desired bin size. Let
    :math:`b_i` and :math:`b_j` denote the binned spike trains and
    :math:`\mu_i` and :math:`\mu_j` their respective averages. Then

    .. math::
         C[i,j] = <b_i-\mu_i, b_j-\mu_j> / (L-1)

    where `<., .>` is the scalar product of two vectors, and :math:`L` is the
    number of bins.

    For an input of N spike trains, an N x N matrix is returned containing the
    covariances for each combination of input spike trains.

    If binary is True, the binned spike trains are clipped to 0 or 1 before
    computing the covariance, so that the binned vectors :math:`b_i` and
    :math:`b_j` are binary.

    Parameters
    ----------
    binned_spiketrain : (N, ) elephant.conversion.BinnedSpikeTrain
        A binned spike train containing the spike trains to be evaluated.
    binary : bool, optional
        If True, the spikes of a particular spike train falling in the same bin
        are counted as 1, resulting in binary binned vectors :math:`b_i`. If
        False, the binned vectors :math:`b_i` contain the spike counts per bin.
        Default: False
    fast : bool, optional
        If `fast=True` and the sparsity of `binned_spiketrain` is `> 0.1`, use
        `np.cov()`. Otherwise, use memory efficient implementation.
        See Notes [2].
        Default: True

    Returns
    -------
    C : (N, N) np.ndarray
        The square matrix of covariances. The element :math:`C[i,j]=C[j,i]` is
        the covariance between `binned_spiketrain[i]` and
        `binned_spiketrain[j]`.

    Raises
    ------
    MemoryError
        When using `fast=True` and `binned_spiketrain` shape is large.

    Warns
    -----
    UserWarning
        If at least one row in `binned_spiketrain` is empty (has no spikes).

    See Also
    --------
    correlation_coefficient : Pearson correlation coefficient

    Notes
    -----
    1. The spike trains in the binned structure are assumed to cover the
       complete time span `[t_start, t_stop)` of `binned_spiketrain`.
    2. Using `fast=True` might lead to `MemoryError`. If it's the case,
       switch to `fast=False`.

    Examples
    --------
    Covariance matrix of two Poisson spike train processes.

    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.spike_train_generation import StationaryPoissonProcess
    >>> from elephant.conversion import BinnedSpikeTrain
    >>> from elephant.spike_train_correlation import covariance

    >>> np.random.seed(1)
    >>> st1 = StationaryPoissonProcess(rate=10*pq.Hz,
    ... t_stop=10.0*pq.s).generate_spiketrain()
    >>> st2 = StationaryPoissonProcess(rate=10*pq.Hz,
    ... t_stop=10.0*pq.s).generate_spiketrain()
    >>> cov_matrix = covariance(BinnedSpikeTrain([st1, st2], bin_size=5*pq.ms))
    >>> cov_matrix # doctest: +SKIP
    array([[ 0.05432316, -0.00152276],
       [-0.00152276,  0.04917234]])


    """
    if binary:
        binned_spiketrain = binned_spiketrain.binarize()

    if fast and binned_spiketrain.sparsity > _SPARSITY_MEMORY_EFFICIENT_THR:
        array = binned_spiketrain.to_array()
        return np.cov(array)

    return _covariance_sparse(
        binned_spiketrain, corrcoef_norm=False)


def correlation_coefficient(binned_spiketrain, binary=False, fast=True):
    r"""
    Calculate the NxN matrix of pairwise Pearson's correlation coefficients
    between all combinations of N binned spike trains.

    For each pair of spike trains :math:`(i,j)`, the correlation coefficient
    :math:`C[i,j]` is obtained by binning :math:`i` and :math:`j` at the
    desired bin size. Let :math:`b_i` and :math:`b_j` denote the binned spike
    trains and :math:`\mu_i` and :math:`\mu_j` their respective means. Then

    .. math::
         C[i,j] = <b_i-\mu_i, b_j-\mu_j> /
                  \sqrt{<b_i-\mu_i, b_i-\mu_i> \cdot <b_j-\mu_j, b_j-\mu_j>}

    where `<., .>` is the scalar product of two vectors.

    For an input of N spike trains, an N x N matrix is returned.
    Each entry in the matrix is a real number ranging between -1 (perfectly
    anti-correlated spike trains) and +1 (perfectly correlated spike trains).
    However, if k-th spike train is empty, k-th row and k-th column of the
    returned matrix are set to np.nan.

    If binary is True, the binned spike trains are clipped to 0 or 1 before
    computing the correlation coefficients, so that the binned vectors
    :math:`b_i` and :math:`b_j` are binary.

    Visualization of this function is covered in Viziphant:
    :func:`viziphant.spike_train_correlation.plot_corrcoef`.


    Parameters
    ----------
    binned_spiketrain : (N, ) elephant.conversion.BinnedSpikeTrain
        A binned spike train containing the spike trains to be evaluated.
    binary : bool, optional
        If True, two spikes of a particular spike train falling in the same bin
        are counted as 1, resulting in binary binned vectors :math:`b_i`. If
        False, the binned vectors :math:`b_i` contain the spike counts per bin.
        Default: False
    fast : bool, optional
        If `fast=True` and the sparsity of `binned_spiketrain` is `> 0.1`, use
        `np.corrcoef()`. Otherwise, use memory efficient implementation.
        See Notes[2]
        Default: True

    Returns
    -------
    C : (N, N) np.ndarray
        The square matrix of correlation coefficients. The element
        :math:`C[i,j]=C[j,i]` is the Pearson's correlation coefficient between
        `binned_spiketrain[i]` and `binned_spiketrain[j]`.
        If `binned_spiketrain` contains only one `neo.SpikeTrain`, C=1.0.

    Raises
    ------
    MemoryError
        When using `fast=True` and `binned_spiketrain` shape is large.

    Warns
    -----
    UserWarning
        If at least one row in `binned_spiketrain` is empty (has no spikes).

    See Also
    --------
    covariance

    Notes
    -----
    1. The spike trains in the binned structure are assumed to cover the
       complete time span `[t_start, t_stop)` of `binned_spiketrain`.
    2. Using `fast=True` might lead to `MemoryError`. If it's the case,
       switch to `fast=False`.

    Examples
    --------
    Correlation coefficient of two Poisson spike train processes.

    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.spike_train_generation import StationaryPoissonProcess
    >>> from elephant.conversion import BinnedSpikeTrain
    >>> from elephant.spike_train_correlation import correlation_coefficient

    >>> np.random.seed(1)
    >>> st1 = StationaryPoissonProcess(rate=10*pq.Hz,
    ... t_stop=10.0*pq.s).generate_spiketrain()
    >>> st2 = StationaryPoissonProcess(rate=10*pq.Hz,
    ... t_stop=10.0*pq.s).generate_spiketrain()
    >>> corrcoef = correlation_coefficient(BinnedSpikeTrain([st1, st2],
    ...     bin_size=5*pq.ms))
    >>> corrcoef # doctest: +SKIP
    array([[ 1.        , -0.02946313],
           [-0.02946313,  1.        ]])

    """
    if binary:
        binned_spiketrain = binned_spiketrain.binarize()

    if fast and binned_spiketrain.sparsity > _SPARSITY_MEMORY_EFFICIENT_THR:
        array = binned_spiketrain.to_array()
        return np.corrcoef(array)

    return _covariance_sparse(
        binned_spiketrain, corrcoef_norm=True)


def corrcoef(*args, **kwargs):
    warnings.warn("'corrcoef' is deprecated; use 'correlation_coefficient'",
                  DeprecationWarning)
    return correlation_coefficient(*args, **kwargs)


def _covariance_sparse(binned_spiketrain, corrcoef_norm):
    r"""
    Memory efficient helper function for `covariance()` and `corrcoef()`
    that performs the complete calculation for either the covariance
    (`corrcoef_norm=False`) or correlation coefficient (`corrcoef_norm=True`).
    Both calculations differ only by the denominator.

    For any two `BinnedSpikeTrain`s :math:`\hat{b_x}` and :math:`\hat{b_y}`
    with mean :math:`\vec{\mu_x}` and :math:`\vec{mu_y}` respectively
    computes the dot product

    .. math::
        <\hat{b_x} - \vec{\mu_x}, \hat{b_y} - \vec{\mu_y}>_{ij} =
            (\hat{b_x} \cdot \hat{b_y}^T)_{ij} -
            \frac{(\vec{N_x}^T \cdot \vec{N_y})_{ij}}{L}

    where :math:`N_x^i = \sum_j{b_x^{ij}}` - the number of spikes in `i`th row
    of :math:`\hat{b_x}`, :math:`L` - the number of bins, and
    :math:`\vec{\mu_x} = \frac{\vec{N_x}}{L}`.

    Parameters
    ----------
    binned_spiketrain : (N, ) elephant.conversion.BinnedSpikeTrain
        See `covariance()` or `corrcoef()`, respectively.
    corrcoef_norm : bool
        Use normalization factor for the correlation coefficient rather than
        for the covariance.

    Warns
    --------
    UserWarning
        If at least one row in `binned_spiketrain` is empty (has no spikes).

    Returns
    -------
    (N, N) np.ndarray
        Pearson correlation or covariance matrix.
    """
    spmat = binned_spiketrain.sparse_matrix
    n_bins = binned_spiketrain.n_bins

    # Check for empty spike trains
    n_spikes_per_row = spmat.sum(axis=1)
    if n_spikes_per_row.min() == 0:
        warnings.warn(
            'Detected empty spike trains (rows) in the binned_spiketrain.')

    res = spmat.dot(spmat.T) - n_spikes_per_row * n_spikes_per_row.T / n_bins
    res = np.asarray(res)
    if corrcoef_norm:
        stdx = np.sqrt(res.diagonal())
        stdx = np.expand_dims(stdx, axis=0)
        res /= (stdx.T * stdx)
    else:
        res /= (n_bins - 1)
    res = np.squeeze(res)
    return res


def cross_correlation_histogram(
        binned_spiketrain_i, binned_spiketrain_j, window='full',
        border_correction=False, binary=False, kernel=None, method='speed',
        cross_correlation_coefficient=False):
    """
    Computes the cross-correlation histogram (CCH) between two binned spike
    trains `binned_spiketrain_i` and `binned_spiketrain_j`.
    :cite:`correlation-Eggermont2010_77`

    Visualization of this function is covered in Viziphant:
    :func:`viziphant.spike_train_correlation.plot_cross_correlation_histogram`.

    Parameters
    ----------
    binned_spiketrain_i, binned_spiketrain_j : BinnedSpikeTrain
        Binned spike trains of lengths N and M to cross-correlate - the output
        of :class:`elephant.conversion.BinnedSpikeTrain`. The input
        spike trains can have any `t_start` and `t_stop`.
    window : {'valid', 'full'} or list of int, optional
        ‘full’: This returns the cross-correlation at each point of overlap,
              with an output shape of (N+M-1,). At the end-points of the
              cross-correlogram, the signals do not overlap completely, and
              boundary effects may be seen.
        ‘valid’: Mode valid returns output of length max(M, N) - min(M, N) + 1.
              The cross-correlation product is only given for points where
              the signals overlap completely.
              Values outside the signal boundary have no effect.
        List of integers (min_lag, max_lag):
              The entries of window are two integers representing the left and
              right extremes (expressed as number of bins) where the
              cross-correlation is computed.
    Default: 'full'
    border_correction : bool, optional
        whether to correct for the border effect. If True, the value of the
        CCH at bin :math:`b` (for :math:`b=-H,-H+1, ...,H`, where :math:`H` is
        the CCH half-length) is multiplied by the correction factor:

        .. math::
                            (H+1)/(H+1-|b|),

        which linearly corrects for loss of bins at the edges.
        Default: False
    binary : bool, optional
        If True, spikes falling in the same bin are counted as a single spike;
        otherwise they are counted as different spikes.
        Default: False
    kernel : np.ndarray or None, optional
        A one dimensional array containing a smoothing kernel applied
        to the resulting CCH. The length N of the kernel indicates the
        smoothing window. The smoothing window cannot be larger than the
        maximum lag of the CCH. The kernel is normalized to unit area before
        being applied to the resulting CCH. Popular choices for the kernel are

          * normalized boxcar kernel: `numpy.ones(N)`
          * hamming: `numpy.hamming(N)`
          * hanning: `numpy.hanning(N)`
          * bartlett: `numpy.bartlett(N)`

        If None, the CCH is not smoothed.
        Default: None
    method : {'speed', 'memory'}, optional
        Defines the algorithm to use. "speed" uses `numpy.correlate` to
        calculate the correlation between two binned spike trains using a
        non-sparse data representation. Due to various optimizations, it is the
        fastest realization. In contrast, the option "memory" uses an own
        implementation to calculate the correlation based on sparse matrices,
        which is more memory efficient but slower than the "speed" option.
        Default: "speed"
    cross_correlation_coefficient : bool, optional
        If True, a normalization is applied to the CCH to obtain the
        cross-correlation  coefficient function ranging from -1 to 1 according
        to Equation (5.10) in :cite:`correlation-Eggermont2010_77`. See Notes.
        Default: False

    Returns
    -------
    cch_result : neo.AnalogSignal
        Containing the cross-correlation histogram between
        `binned_spiketrain_i` and `binned_spiketrain_j`.

        Offset bins correspond to correlations at delays equivalent
        to the differences between the spike times of `binned_spiketrain_i` and
        those of `binned_spiketrain_j`: an entry at positive lag corresponds to
        a spike in `binned_spiketrain_j` following a spike in
        `binned_spiketrain_i` bins to the right, and an entry at negative lag
        corresponds to a spike in `binned_spiketrain_i` following a spike in
        `binned_spiketrain_j`.

        To illustrate this definition, consider two spike trains with the same
        `t_start` and `t_stop`:
        `binned_spiketrain_i` ('reference neuron') : 0 0 0 0 1 0 0 0 0 0 0
        `binned_spiketrain_j` ('target neuron')    : 0 0 0 0 0 0 0 1 0 0 0
        Here, the CCH will have an entry of `1` at `lag=+3`.

        Consistent with the definition of `neo.AnalogSignals`, the time axis
        represents the left bin borders of each histogram bin. For example,
        the time axis might be:
        `np.array([-2.5 -1.5 -0.5 0.5 1.5]) * ms`
    lags : np.ndarray
        Contains the IDs of the individual histogram bins, where the central
        bin has ID 0, bins to the left have negative IDs and bins to the right
        have positive IDs, e.g.,:
        `np.array([-3, -2, -1, 0, 1, 2, 3])`

    Notes
    -----
    1. The Eq. (5.10) in :cite:`correlation-Eggermont2010_77` is valid for
       binned spike trains with at most one spike per bin. For a general case,
       refer to the implementation of `_covariance_sparse()`.
    2. Alias: `cch`

    Examples
    --------
    Plot the cross-correlation histogram between two Poisson spike trains

    >>> import elephant
    >>> import quantities as pq
    >>> import numpy as np
    >>> from elephant.conversion import BinnedSpikeTrain
    >>> from elephant.spike_train_generation import StationaryPoissonProcess
    >>> from elephant.spike_train_correlation import cross_correlation_histogram  # noqa
    >>> np.random.seed(1)
    >>> binned_spiketrain_i = BinnedSpikeTrain(
    ...        StationaryPoissonProcess(
    ...            10. * pq.Hz, t_start=0 * pq.ms,
    ...             t_stop=5000 * pq.ms).generate_spiketrain(),
    ...        bin_size=5. * pq.ms)
    >>> binned_spiketrain_j = BinnedSpikeTrain(
    ...        StationaryPoissonProcess(
    ...            10. * pq.Hz, t_start=0 * pq.ms,
    ...             t_stop=5000 * pq.ms).generate_spiketrain(),
    ...        bin_size=5. * pq.ms)

    >>> cc_hist, lags = cross_correlation_histogram(
    ...        binned_spiketrain_i, binned_spiketrain_j, window=[-10, 10],
    ...        border_correction=False,
    ...        binary=False, kernel=None)
    >>> print(cc_hist.flatten()) # doctest: +SKIP
    [ 5.  3.  3.  2.  4.  0.  1.  5.  3.  4.  2.  2.  2.  5.
      1.  2.  4.  2. -0.  3.  3.] dimensionless

    >>> lags # doctest: +SKIP
    array([-10,  -9,  -8,  -7,  -6,  -5,  -4,  -3,  -2,  -1,
         0,   1,   2,   3,   4,   5,   6,   7,   8,   9,
        10], dtype=int32)


    """

    # Check that the spike trains are binned with the same temporal
    # resolution
    if binned_spiketrain_i.shape[0] != 1 or \
            binned_spiketrain_j.shape[0] != 1:
        raise ValueError("Spike trains must be one dimensional")

    # rescale to the common units
    # this does not change the data - only its representation
    binned_spiketrain_j.rescale(binned_spiketrain_i.units)

    if not np.isclose(binned_spiketrain_i._bin_size,
                      binned_spiketrain_j._bin_size):
        raise ValueError("Bin sizes must be equal")

    bin_size = binned_spiketrain_i._bin_size
    left_edge_min = -binned_spiketrain_i.n_bins + 1
    right_edge_max = binned_spiketrain_j.n_bins - 1

    t_lags_shift = (binned_spiketrain_j._t_start -
                    binned_spiketrain_i._t_start) / bin_size
    if not np.isclose(t_lags_shift, round(t_lags_shift)):
        # For example, if bin_size=1 ms, binned_spiketrain_i.t_start=0 ms, and
        # binned_spiketrain_j.t_start=0.5 ms then there is a global shift in
        # the binning of the spike trains.
        raise ValueError(
            "Binned spiketrains time shift is not multiple of bin_size")
    t_lags_shift = int(round(t_lags_shift))

    # In the examples below we fix st2 and "move" st1.
    # Zero-lag is equal to `max(st1.t_start, st2.t_start)`.
    # Binned spiketrains (t_start and t_stop) with bin_size=1ms:
    # 1) st1=[3, 8] ms, st2=[1, 13] ms
    #    t_start_shift = -2 ms
    #    zero-lag is at 3 ms
    # 2) st1=[1, 7] ms, st2=[2, 9] ms
    #    t_start_shift = 1 ms
    #    zero-lag is at 2 ms
    # 3) st1=[1, 7] ms, st2=[4, 6] ms
    #    t_start_shift = 3 ms
    #    zero-lag is at 4 ms

    # Find left and right edges of unaligned (time-dropped) time signals
    if len(window) == 2 and np.issubdtype(type(window[0]), np.integer) \
            and np.issubdtype(type(window[1]), np.integer):
        # ex. 1) lags range: [w[0] - 2, w[1] - 2] ms
        # ex. 2) lags range: [w[0] + 1, w[1] + 1] ms
        # ex. 3) lags range: [w[0] + 3, w[0] + 3] ms
        if window[0] >= window[1]:
            raise ValueError(
                "Window's left edge ({left}) must be lower than the right "
                "edge ({right})".format(left=window[0], right=window[1]))
        left_edge, right_edge = np.subtract(window, t_lags_shift)
        if left_edge < left_edge_min or right_edge > right_edge_max:
            raise ValueError(
                "The window exceeds the length of the spike trains")
        lags = np.arange(window[0], window[1] + 1, dtype=np.int32)
        cch_mode = 'pad'
    elif window == 'full':
        # cch computed for all the possible entries
        # ex. 1) lags range: [-6, 9] ms
        # ex. 2) lags range: [-4, 7] ms
        # ex. 3) lags range: [-2, 4] ms
        left_edge = left_edge_min
        right_edge = right_edge_max
        lags = np.arange(left_edge + t_lags_shift,
                         right_edge + 1 + t_lags_shift, dtype=np.int32)
        cch_mode = window
    elif window == 'valid':
        lags = _CrossCorrHist.get_valid_lags(binned_spiketrain_i,
                                             binned_spiketrain_j)
        left_edge, right_edge = lags[(0, -1), ]
        cch_mode = window
    else:
        raise ValueError("Invalid window parameter")

    if binary:
        binned_spiketrain_i = binned_spiketrain_i.binarize()
        binned_spiketrain_j = binned_spiketrain_j.binarize()

    cch_builder = _CrossCorrHist(binned_spiketrain_i, binned_spiketrain_j,
                                 window=(left_edge, right_edge))
    if method == 'memory':
        cross_corr = cch_builder.correlate_memory(cch_mode=cch_mode)
    else:
        cross_corr = cch_builder.correlate_speed(cch_mode=cch_mode)

    if border_correction:
        if window == 'valid':
            warnings.warn(
                "Border correction does not have any effect in "
                "'valid' window mode since there are no border effects!")
        else:
            cross_corr = cch_builder.border_correction(cross_corr)
    if kernel is not None:
        cross_corr = cch_builder.kernel_smoothing(cross_corr, kernel=kernel)
    if cross_correlation_coefficient:
        cross_corr = cch_builder.cross_correlation_coefficient(cross_corr)

    normalization = 'normalized' if cross_correlation_coefficient else 'counts'
    annotations = dict(window=window, border_correction=border_correction,
                       binary=binary, kernel=kernel is not None,
                       normalization=normalization)
    annotations = dict(cch_parameters=annotations)

    # Transform the array count into an AnalogSignal
    t_start = pq.Quantity((lags[0] - 0.5) * bin_size,
                          units=binned_spiketrain_i.units, copy=False)
    cch_result = neo.AnalogSignal(
        signal=np.expand_dims(cross_corr, axis=1),
        units=pq.dimensionless,
        t_start=t_start,
        sampling_period=binned_spiketrain_i.bin_size,
        **annotations)
    return cch_result, lags


# Alias for common abbreviation
cch = cross_correlation_histogram


def spike_time_tiling_coefficient(spiketrain_i: neo.core.SpikeTrain,
                                  spiketrain_j: neo.core.SpikeTrain,
                                  dt: pq.Quantity = 0.005 * pq.s) -> float:
    """
    Calculates the Spike Time Tiling Coefficient (STTC) as described in
    :cite:`correlation-Cutts2014_14288` following their implementation in C.
    The STTC is a pairwise measure of correlation between spike trains.
    It has been proposed as a replacement for the correlation index as it
    presents several advantages (e.g. it's not confounded by firing rate,
    appropriately distinguishes lack of correlation from anti-correlation,
    periods of silence don't add to the correlation, and it's sensitive to
    firing patterns).

    The STTC is calculated as follows:

    .. math::
        STTC = 1/2((PA - TB)/(1 - PA*TB) + (PB - TA)/(1 - PB*TA))

    Where `PA` is the proportion of spikes from train 1 that lie within
    `[-dt, +dt]` of any spike of train 2 divided by the total number of spikes
    in train 1, `PB` is the same proportion for the spikes in train 2;
    `TA` is the proportion of total recording time within `[-dt, +dt]` of any
    spike in train 1, TB is the same proportion for train 2.
    For :math:`TA = PB = 1` and for :math:`TB = PA = 1`
    the resulting :math:`0/0` is replaced with :math:`1`,
    since every spike from the train with :math:`T = 1` is within
    `[-dt, +dt]` of a spike of the other train.

    This is a Python implementation compatible with the elephant library of
    the original code by C. Cutts written in C and available `here
    <https://github.com/CCutts/Detecting_pairwise_correlations_in_spike_trains/
    blob/master/spike_time_tiling_coefficient.c>`_:

    Parameters
    ----------
    spiketrain_i, spiketrain_j : :class:`neo.core.SpikeTrain`
        Spike trains to cross-correlate. They must have the same `t_start` and
        `t_stop`.
    dt : pq.Quantity.
        The synchronicity window is used for both: the quantification of the
        proportion of total recording time that lies `[-dt, +dt]` of each spike
        in each train and the proportion of spikes in `spiketrain_i` that lies
        `[-dt, +dt]` of any spike in `spiketrain_j`.
        Default : `0.005 * pq.s`

    Returns
    -------
    index : :class:`float` or :obj:`numpy.nan`
        The spike time tiling coefficient (STTC). Returns :obj:`numpy.nan` if
        any spike train is empty.

    Notes
    -----
    Alias: `sttc`

    Examples
    --------
    >>> import neo
    >>> import quantities as pq
    >>> from elephant.spike_train_correlation import spike_time_tiling_coefficient  # noqa

    >>> spiketrain1 = neo.SpikeTrain([1.3, 7.56, 15.87, 28.23, 30.9, 34.2,
    ...     38.2, 43.2], units='ms', t_stop=50)
    >>> spiketrain2 = neo.SpikeTrain([1.02, 2.71, 18.82, 28.46, 28.79, 43.6],
    ...     units='ms', t_stop=50)
    >>> spike_time_tiling_coefficient(spiketrain1, spiketrain2)
    0.4958601655933762

    """
    # input checks
    if dt <= 0 * pq.s:
        raise ValueError(f"dt must be > 0, found: {dt}")

    check_neo_consistency([spiketrain_j, spiketrain_i], neo.core.SpikeTrain)

    if dt.units != spiketrain_i.units:
        dt = dt.rescale(spiketrain_i.units)

    def run_p(spiketrain_j: neo.core.SpikeTrain,
              spiketrain_i: neo.core.SpikeTrain,
              dt: pq.Quantity = dt) -> float:
        """
        Returns number of spikes in spiketrain_j which lie within +- dt of
        any spike from spiketrain_i, divided by the total number of spikes in
        spiketrain_j
        """
        # Create a boolean array where each element represents whether a spike
        # in spiketrain_j lies within +- dt of any spike in spiketrain_i.
        tiled_spikes_j = np.isclose(
            spiketrain_j.times.magnitude[:, np.newaxis],
            spiketrain_i.times.magnitude,
            atol=dt.item())
        # Determine which spikes in spiketrain_j satisfy the time window
        # condition.
        tiled_spike_indices = np.any(tiled_spikes_j, axis=1)
        # Extract the spike times in spiketrain_j that satisfy the condition.
        tiled_spikes_j = spiketrain_j[tiled_spike_indices]
        # Calculate the ratio of matching spikes in j to the total spikes in j.
        return len(tiled_spikes_j)/len(spiketrain_j)

    def run_t(spiketrain: neo.core.SpikeTrain, dt: pq.Quantity = dt) -> float:
        """
        Calculate the proportion of the total recording time 'tiled' by spikes.
        """
        # Get the numerical value of 'dt'.
        dt = dt.item()
        # Get the start and stop times of the spike train.
        t_start = spiketrain.t_start.item()
        t_stop = spiketrain.t_stop.item()
        # Get the spike times as a NumPy array.
        sorted_spikes = spiketrain.times.magnitude
        # Check if spikes are sorted and sort them if not.
        if (np.diff(sorted_spikes) < 0).any():
            sorted_spikes = np.sort(sorted_spikes)

        # Calculate the time differences between consecutive spikes.
        diff_spikes = np.diff(sorted_spikes)
        # Calculate durations of spike overlaps within a time window of 2 * dt.
        overlap_durations = diff_spikes[diff_spikes <= 2 * dt]
        covered_time_overlap = np.sum(overlap_durations)

        # Calculate the durations of non-overlapping spikes.
        non_overlap_durations = diff_spikes[diff_spikes > 2 * dt]
        covered_time_non_overlap = len(non_overlap_durations) * 2 * dt

        # Check if the first and last spikes are within +/-dt of the start
        # and end.
        # If so, adjust the overlapping and non-overlapping times accordingly.
        if sorted_spikes[0] - t_start < dt:
            covered_time_overlap += sorted_spikes[0] - t_start
        else:
            covered_time_non_overlap += dt
        if t_stop - sorted_spikes[- 1] < dt:
            covered_time_overlap += t_stop - sorted_spikes[-1]
        else:
            covered_time_non_overlap += dt

        # Calculate the total time covered by spikes and the total recording
        # time.
        total_time_covered = covered_time_overlap + covered_time_non_overlap
        total_time = t_stop - t_start
        # Calculate and return the proportion of the total recording time
        # covered by spikes.
        return total_time_covered / total_time

    if len(spiketrain_i) == 0 or len(spiketrain_j) == 0:
        index = np.nan
    else:
        TA = run_t(spiketrain_j, dt)
        TB = run_t(spiketrain_i, dt)
        PA = run_p(spiketrain_j, spiketrain_i, dt)
        PB = run_p(spiketrain_i, spiketrain_j, dt)

        # check if the P and T values are 1 to avoid division by zero
        # This only happens for TA = PB = 1 and/or TB = PA = 1,
        # which leads to 0/0 in the calculation of the index.
        # In those cases, every spike in the train with P = 1
        # is within dt of a spike in the other train,
        # so we set the respective (partial) index to 1.
        if PA * TB == 1 and PB * TA == 1:
            index = 1.
        elif PA * TB == 1:
            index = 0.5 + 0.5 * (PB - TA) / (1 - PB * TA)
        elif PB * TA == 1:
            index = 0.5 + 0.5 * (PA - TB) / (1 - PA * TB)
        else:
            index = 0.5 * (PA - TB) / (1 - PA * TB) + \
                    0.5 * (PB - TA) / (1 - PB * TA)
    return index


sttc = spike_time_tiling_coefficient


def spike_train_timescale(binned_spiketrain, max_tau):
    r"""
    Calculates the auto-correlation time of a binned spike train; uses the
    definition of the auto-correlation time proposed in
    :cite:`correlation-Wieland2015_040901` (Eq. 6):

    .. math::
        \tau_\mathrm{corr} = \int_{-\tau_\mathrm{max}}^{\tau_\mathrm{max}}\
            \left[ \frac{\hat{C}(\tau)}{\hat{C}(0)} \right]^2 d\tau

    where :math:`\hat{C}(\tau) = C(\tau)-\nu\delta(\tau)` denotes
    the auto-correlation function excluding the Dirac delta at zero timelag.

    Parameters
    ----------
    binned_spiketrain : elephant.conversion.BinnedSpikeTrain
        A binned spike train containing the spike train to be evaluated.
    max_tau : pq.Quantity
        Maximal integration time :math:`\tau_{max}` of the auto-correlation
        function. It needs to be a multiple of the `bin_size` of
        `binned_spiketrain`.

    Returns
    -------
    timescale : pq.Quantity
        The auto-correlation time of the binned spiketrain with the same units
        as in the input. If `binned_spiketrain` has less than 2 spikes, a
        warning is raised and `np.nan` is returned.

    Notes
    -----
    * :math:`\tau_\mathrm{max}` is a critical parameter: numerical estimates
      of the auto-correlation functions are inherently noisy. Due to the
      square in the definition above, this noise is integrated. Thus, it is
      necessary to introduce a cutoff for the numerical integration - this
      cutoff should be neither smaller than the true auto-correlation time
      nor much bigger.
    * The bin size of `binned_spiketrain` is another critical parameter as it
      defines the discretization of the integral :math:`d\tau`. If it is too
      big, the numerical approximation of the integral is inaccurate.

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.spike_train_correlation import spike_train_timescale
    >>> from elephant.conversion import BinnedSpikeTrain
    >>> spiketrain = neo.SpikeTrain([1, 5, 7, 8], units='ms', t_stop=10*pq.ms)
    >>> bst = BinnedSpikeTrain(spiketrain, bin_size=1 * pq.ms)
    >>> spike_train_timescale(bst, max_tau=5 * pq.ms)
    array(14.11111111) * ms

    """
    if binned_spiketrain.get_num_of_spikes() < 2:
        warnings.warn("Spike train contains less than 2 spikes! "
                      "np.nan will be returned.")
        return np.nan

    bin_size = binned_spiketrain._bin_size
    try:
        max_tau = max_tau.rescale(binned_spiketrain.units).item()
    except (AttributeError, ValueError):
        raise ValueError("max_tau needs units of time")

    # safe casting of max_tau/bin_size to integer
    max_tau_bins = int(round(max_tau / bin_size))
    if not np.isclose(max_tau, max_tau_bins * bin_size):
        raise ValueError("max_tau has to be a multiple of the bin_size")

    cch_window = [-max_tau_bins, max_tau_bins]
    corrfct, bin_ids = cross_correlation_histogram(
        binned_spiketrain, binned_spiketrain, window=cch_window,
        cross_correlation_coefficient=True
    )
    # Take only t > 0 values, in particular neglecting the delta peak.
    start_id = corrfct.time_index((bin_size / 2) * binned_spiketrain.units)
    corrfct = corrfct.magnitude.squeeze()[start_id:]

    # Calculate the timescale using trapezoidal integration
    integr = (corrfct / corrfct[0]) ** 2
    timescale = 2 * integrate.trapezoid(integr, dx=bin_size)
    return pq.Quantity(timescale, units=binned_spiketrain.units, copy=False)



"""
Functions for analysing and estimating firing patterns and connectivity among
neurons in order to better understand the underlying neural networks and
information flow between neurons.


Network connectivity estimation
*******************************

.. autosummary::
    :toctree: _toctree/functional_connectivity/

    total_spiking_probability_edges

References
----------

.. bibliography::
   :keyprefix: functional_connectivity-


:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from elephant.functional_connectivity_src.total_spiking_probability_edges \
    import (
        total_spiking_probability_edges,
    )

__all__ = ["total_spiking_probability_edges"]



# -*- coding: utf-8 -*-
"""
Unit tests for elephant.utils
"""

import unittest

import neo
import numpy as np
import quantities as pq

from elephant import utils
from numpy.testing import assert_array_equal

from elephant.spike_train_generation import StationaryPoissonProcess
from elephant.trials import TrialsFromLists
from neo.core.spiketrainlist import SpikeTrainList
from neo.core import SpikeTrain


class TestUtils(unittest.TestCase):

    def test_check_neo_consistency(self):
        self.assertRaises(TypeError,
                          utils.check_neo_consistency,
                          [], object_type=neo.SpikeTrain)
        self.assertRaises(TypeError,
                          utils.check_neo_consistency,
                          [neo.SpikeTrain([1]*pq.s, t_stop=2*pq.s),
                           np.arange(2)], object_type=neo.SpikeTrain)
        self.assertRaises(ValueError,
                          utils.check_neo_consistency,
                          [neo.SpikeTrain([1]*pq.s,
                                          t_start=1*pq.s,
                                          t_stop=2*pq.s),
                           neo.SpikeTrain([1]*pq.s,
                                          t_start=0*pq.s,
                                          t_stop=2*pq.s)],
                          object_type=neo.SpikeTrain)
        self.assertRaises(ValueError,
                          utils.check_neo_consistency,
                          [neo.SpikeTrain([1]*pq.s, t_stop=2*pq.s),
                           neo.SpikeTrain([1]*pq.s, t_stop=3*pq.s)],
                          object_type=neo.SpikeTrain)
        self.assertRaises(ValueError,
                          utils.check_neo_consistency,
                          [neo.SpikeTrain([1]*pq.ms, t_stop=2000*pq.ms),
                           neo.SpikeTrain([1]*pq.s, t_stop=2*pq.s)],
                          object_type=neo.SpikeTrain)

    def test_round_binning_errors(self):
        n_bins = utils.round_binning_errors(0.999999, tolerance=1e-6)
        self.assertEqual(n_bins, 1)
        self.assertEqual(utils.round_binning_errors(0.999999, tolerance=None),
                         0)
        array = np.array([0, 0.7, 1 - 1e-8, 1 - 1e-9])
        corrected = utils.round_binning_errors(array.copy())
        assert_array_equal(corrected, [0, 0, 1, 1])
        assert_array_equal(
            utils.round_binning_errors(array.copy(), tolerance=None),
            [0, 0, 0, 0])


class DecoratorTest:
    """
    This class is used as a mock for testing the decorator.
    """
    @utils.trials_to_list_of_spiketrainlist
    def method_to_decorate(self, trials=None, trials_obj=None):
        # This is just a mock implementation for testing purposes
        if trials_obj:
            return trials_obj
        else:
            return trials


class TestTrialsToListOfSpiketrainlist(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.n_channels = 10
        cls.n_trials = 5
        cls.list_of_list_of_spiketrains = [
            StationaryPoissonProcess(rate=5 * pq.Hz, t_stop=1000.0 * pq.ms
                                     ).generate_n_spiketrains(cls.n_channels)
            for _ in range(cls.n_trials)]
        cls.trial_object = TrialsFromLists(cls.list_of_list_of_spiketrains)

    def test_decorator_applied(self):
        # Test that the decorator is applied correctly
        self.assertTrue(hasattr(
            DecoratorTest.method_to_decorate, '__wrapped__'
            ))

    def test_decorator_return_with_trials_input_as_arg(self):
        # Test if decorator takes in trial-object and returns
        # list of spiketrainlists
        new_class = DecoratorTest()
        list_of_spiketrainlists = new_class.method_to_decorate(
            self.trial_object)
        self.assertEqual(len(list_of_spiketrainlists), self.n_trials)
        for spiketrainlist in list_of_spiketrainlists:
            self.assertIsInstance(spiketrainlist, SpikeTrainList)

    def test_decorator_return_with_list_of_lists_input_as_arg(self):
        # Test if decorator takes in list of lists of spiketrains
        # and does not change input
        new_class = DecoratorTest()
        list_of_list_of_spiketrains = new_class.method_to_decorate(
            self.list_of_list_of_spiketrains)
        self.assertEqual(len(list_of_list_of_spiketrains), self.n_trials)
        for list_of_spiketrains in list_of_list_of_spiketrains:
            self.assertIsInstance(list_of_spiketrains, list)
            for spiketrain in list_of_spiketrains:
                self.assertIsInstance(spiketrain, SpikeTrain)

    def test_decorator_return_with_trials_input_as_kwarg(self):
        # Test if decorator takes in trial-object and returns
        # list of spiketrainlists
        new_class = DecoratorTest()
        list_of_spiketrainlists = new_class.method_to_decorate(
            trials_obj=self.trial_object)
        self.assertEqual(len(list_of_spiketrainlists), self.n_trials)
        for spiketrainlist in list_of_spiketrainlists:
            self.assertIsInstance(spiketrainlist, SpikeTrainList)

    def test_decorator_return_with_list_of_lists_input_as_kwarg(self):
        # Test if decorator takes in list of lists of spiketrains
        # and does not change input
        new_class = DecoratorTest()
        list_of_list_of_spiketrains = new_class.method_to_decorate(
            trials_obj=self.list_of_list_of_spiketrains)
        self.assertEqual(len(list_of_list_of_spiketrains), self.n_trials)
        for list_of_spiketrains in list_of_list_of_spiketrains:
            self.assertIsInstance(list_of_spiketrains, list)
            for spiketrain in list_of_spiketrains:
                self.assertIsInstance(spiketrain, SpikeTrain)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Unit tests for the trials objects.

:copyright: Copyright 2014-2024 by the Elephant team, see AUTHORS.txt.
:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest
import neo.utils
import quantities as pq
from neo.core import AnalogSignal

from elephant.spike_train_generation import StationaryPoissonProcess
from elephant.trials import TrialsFromBlock, TrialsFromLists


def _create_trials_block(n_trials: int = 0,
                         n_spiketrains: int = 2,
                         n_analogsignals: int = 2) -> neo.core.Block:
    """ Create block with n_trials, n_spiketrains and n_analog_signals """
    block = neo.Block(name='test_block')
    for trial in range(n_trials):
        segment = neo.Segment(name=f'No. {trial}')
        spiketrains = StationaryPoissonProcess(rate=50. * pq.Hz,
                                               t_start=0 * pq.ms,
                                               t_stop=1000 * pq.ms
                                               ).generate_n_spiketrains(
            n_spiketrains=n_spiketrains)
        analogsignals = [AnalogSignal(signal=[.01, 3.3, 9.3], units='uV',
                                      sampling_rate=1 * pq.Hz)
                         for _ in range(n_analogsignals)]
        for spiketrain in spiketrains:
            segment.spiketrains.append(spiketrain)
        for analogsignal in analogsignals:
            segment.analogsignals.append(analogsignal)
        block.segments.append(segment)
    return block


#########
# Tests #
#########


class TrialsFromBlockTestCase(unittest.TestCase):
    """Tests for elephant.trials.TrialsFromBlock class"""

    @classmethod
    def setUpClass(cls) -> None:
        """
        Run once before tests:
        """

        block = _create_trials_block(n_trials=36)
        cls.block = block
        cls.trial_object = TrialsFromBlock(block,
                                           description='trials are segments')

    def setUp(self) -> None:
        """
        Run before every test:
        """

    def test_trials_from_block_description(self) -> None:
        """
        Test description of the trials object.
        """
        self.assertEqual(self.trial_object.description, 'trials are segments')

    def test_trials_from_block_get_item(self) -> None:
        """
        Test get a trial from the trials.
        """
        self.assertIsInstance(self.trial_object[0], neo.core.Segment)

    def test_trials_from_block_get_trial_as_segment(self) -> None:
        """
        Test get a trial from the trials.
        """
        self.assertIsInstance(
            self.trial_object.get_trial_as_segment(0),
            neo.core.Segment)
        self.assertIsInstance(
            self.trial_object.get_trial_as_segment(0).spiketrains[0],
            neo.core.SpikeTrain)
        self.assertIsInstance(
            self.trial_object.get_trial_as_segment(0).analogsignals[0],
            neo.core.AnalogSignal)

    def test_trials_from_block_get_trials_as_block(self) -> None:
        """
        Test get a block from list of trials.
        """
        block = self.trial_object.get_trials_as_block([0, 3, 5])
        self.assertIsInstance(block, neo.core.Block)
        self.assertIsInstance(self.trial_object.get_trials_as_block(),
                              neo.core.Block)
        self.assertEqual(len(block.segments), 3)

    def test_trials_from_block_get_trials_as_list(self) -> None:
        """
        Test get a list of segments from list of trials.
        """
        list_of_trials = self.trial_object.get_trials_as_list([0, 3, 5])
        self.assertIsInstance(list_of_trials, list)
        self.assertIsInstance(self.trial_object.get_trials_as_list(), list)
        self.assertIsInstance(list_of_trials[0], neo.core.Segment)
        self.assertEqual(len(list_of_trials), 3)

    def test_trials_from_block_n_trials(self) -> None:
        """
        Test get number of trials.
        """
        self.assertEqual(self.trial_object.n_trials, len(self.block.segments))

    def test_trials_from_block_n_spiketrains_trial_by_trial(self) -> None:
        """
        Test get number of spiketrains per trial.
        """
        self.assertEqual(self.trial_object.n_spiketrains_trial_by_trial,
                         [len(trial.spiketrains) for trial in
                          self.block.segments])

    def test_trials_from_block_n_analogsignals_trial_by_trial(self) -> None:
        """
        Test get number of analogsignals per trial.
        """
        self.assertEqual(self.trial_object.n_analogsignals_trial_by_trial,
                         [len(trial.analogsignals) for trial in
                          self.block.segments])

    def test_trials_from_block_get_spiketrains_from_trial_as_list(self
                                                                  ) -> None:
        """
        Test get spiketrains from trial as list
        """
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_list(0),
            neo.core.spiketrainlist.SpikeTrainList)
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_list(0)[0],
            neo.core.SpikeTrain)

    def test_trials_from_list_get_spiketrains_from_trial_as_segment(self
                                                                    ) -> None:
        """
        Test get spiketrains from trial as segment
        """
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_segment(0),
            neo.core.Segment)
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_segment(
                0).spiketrains[0], neo.core.SpikeTrain)

    def test_trials_from_block_get_analogsignals_from_trial_as_list(self
                                                                    ) -> None:
        """
        Test get analogsignals from trial as list
        """
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_list(0), list)
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_list(0)[0],
            neo.core.AnalogSignal)

    def test_trials_from_list_get_analogsignals_from_trial_as_segment(self) \
            -> None:
        """
        Test get spiketrains from trial as segment
        """
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_segment(0),
            neo.core.Segment)
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_segment(
                0).analogsignals[0], neo.core.AnalogSignal)


class TrialsFromListTestCase(unittest.TestCase):
    """Tests for elephant.trials.TrialsFromList class"""

    @classmethod
    def setUpClass(cls) -> None:
        """
        Run once before tests:
        Download the dataset from elephant_data
        """
        block = _create_trials_block(n_trials=36)

        # Create Trialobject as list of lists
        # add spiketrains
        trial_list = [[spiketrain for spiketrain in trial.spiketrains]
                      for trial in block.segments]
        # add analogsignals
        for idx, trial in enumerate(block.segments):
            for analogsignal in trial.analogsignals:
                trial_list[idx].append(analogsignal)
        cls.trial_list = trial_list

        cls.trial_object = TrialsFromLists(trial_list,
                                           description='trial is a list')

    def setUp(self) -> None:
        """
        Run before every test:
        """

    def test_trials_from_list_description(self) -> None:
        """
        Test description of the trials object.
        """
        self.assertEqual(self.trial_object.description, 'trial is a list')

    def test_trials_from_list_get_item(self) -> None:
        """
        Test get a trial from the trials.
        """
        self.assertIsInstance(self.trial_object[0],
                              neo.core.Segment)
        self.assertIsInstance(self.trial_object[0].spiketrains[0],
                              neo.core.SpikeTrain)
        self.assertIsInstance(self.trial_object[0].analogsignals[0],
                              neo.core.AnalogSignal)

    def test_trials_from_list_get_trial_as_segment(self) -> None:
        """
        Test get a trial from the trials.
        """
        self.assertIsInstance(
            self.trial_object.get_trial_as_segment(0),
            neo.core.Segment)
        self.assertIsInstance(
            self.trial_object.get_trial_as_segment(0).spiketrains[0],
            neo.core.SpikeTrain)
        self.assertIsInstance(
            self.trial_object.get_trial_as_segment(0).analogsignals[0],
            neo.core.AnalogSignal)

    def test_trials_from_list_get_trials_as_block(self) -> None:
        """
        Test get a block from list of trials.
        """
        block = self.trial_object.get_trials_as_block([0, 3, 5])
        self.assertIsInstance(block, neo.core.Block)
        self.assertIsInstance(self.trial_object.get_trials_as_block(),
                              neo.core.Block)
        self.assertEqual(len(block.segments), 3)

    def test_trials_from_list_get_trials_as_list(self) -> None:
        """
        Test get a list of segments from list of trials.
        """
        list_of_trials = self.trial_object.get_trials_as_list([0, 3, 5])
        self.assertIsInstance(list_of_trials, list)
        self.assertIsInstance(self.trial_object.get_trials_as_list(), list)
        self.assertIsInstance(list_of_trials[0], neo.core.Segment)
        self.assertEqual(len(list_of_trials), 3)

    def test_trials_from_list_n_trials(self) -> None:
        """
        Test get number of trials.
        """
        self.assertEqual(self.trial_object.n_trials, len(self.trial_list))

    def test_trials_from_list_n_spiketrains_trial_by_trial(self) -> None:
        """
        Test get number of spiketrains per trial.
        """
        self.assertEqual(self.trial_object.n_spiketrains_trial_by_trial,
                         [sum(map(lambda x: isinstance(x, neo.core.SpikeTrain),
                                  trial)) for trial in self.trial_list])

    def test_trials_from_list_n_analogsignals_trial_by_trial(self) -> None:
        """
        Test get number of analogsignals per trial.
        """
        self.assertEqual(self.trial_object.n_analogsignals_trial_by_trial,
                         [sum(map(lambda x: isinstance(x,
                                                       neo.core.AnalogSignal),
                                  trial)) for trial in self.trial_list])

    def test_trials_from_list_get_spiketrains_from_trial_as_list(self) -> None:
        """
        Test get spiketrains from trial as list
        """
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_list(0),
            neo.core.spiketrainlist.SpikeTrainList)
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_list(0)[0],
            neo.core.SpikeTrain)

    def test_trials_from_list_get_spiketrains_from_trial_as_segment(self
                                                                    ) -> None:
        """
        Test get spiketrains from trial as segment
        """
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_segment(0),
            neo.core.Segment)
        self.assertIsInstance(
            self.trial_object.get_spiketrains_from_trial_as_segment(
                0).spiketrains[0], neo.core.SpikeTrain)

    def test_trials_from_list_get_analogsignals_from_trial_as_list(self
                                                                   ) -> None:
        """
        Test get analogsignals from trial as list
        """
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_list(0), list)
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_list(0)[0],
            neo.core.AnalogSignal)

    def test_trials_from_list_get_analogsignals_from_trial_as_segment(self
                                                                      ) \
            -> None:
        """
        Test get spiketrains from trial as segment
        """
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_segment(0),
            neo.core.Segment)
        self.assertIsInstance(
            self.trial_object.get_analogsignals_from_trial_as_segment(
                0).analogsignals[0], neo.core.AnalogSignal)


if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path
from typing import Tuple, Union

from neo import SpikeTrain
import numpy as np
from quantities import millisecond as ms
from scipy.io import loadmat

from elephant.conversion import BinnedSpikeTrain
from elephant.functional_connectivity_src.total_spiking_probability_edges \
    import (generate_filter_pairs,
            normalized_cross_correlation,
            TspeFilterPair,
            total_spiking_probability_edges,
            )

from elephant.datasets import download_datasets


class TotalSpikingProbabilityEdgesTestCase(unittest.TestCase):
    def test_generate_filter_pairs(self):
        a = [1]
        b = [1]
        c = [1]
        test_output = [
            TspeFilterPair(
                edge_filter=np.array([-1.0, 0.0, 2.0, 0.0, -1.0]),
                running_total_filter=np.array([1.0]),
                needed_padding=2,
                surrounding_window_size=1,
                observed_window_size=1,
                crossover_window_size=1,
            )
        ]

        function_output = generate_filter_pairs(a, b, c)

        for filter_pair_function, filter_pair_test in zip(function_output,
                                                          test_output):
            np.testing.assert_array_equal(
                filter_pair_function.edge_filter,
                filter_pair_test.edge_filter)

            np.testing.assert_array_equal(
                filter_pair_function.running_total_filter,
                filter_pair_test.running_total_filter)

            self.assertEqual(filter_pair_function.needed_padding,
                             filter_pair_test.needed_padding)

            self.assertEqual(filter_pair_function.surrounding_window_size,
                             filter_pair_test.surrounding_window_size)

            self.assertEqual(filter_pair_function.observed_window_size,
                             filter_pair_test.observed_window_size)

            self.assertEqual(filter_pair_function.crossover_window_size,
                             filter_pair_test.crossover_window_size)

    def test_normalized_cross_correlation(self):
        # Generate Spiketrains
        delay_time = 5
        spike_times = [3, 4, 5] * ms
        spike_times_delayed = spike_times + delay_time * ms

        spiketrains = BinnedSpikeTrain(
            [SpikeTrain(spike_times, t_stop=20.0 * ms),
             SpikeTrain(spike_times_delayed, t_stop=20.0 * ms),],
            bin_size=1 * ms,
        )

        test_output = np.array([[[0.0, 0.0], [1.1, 0.0]], [[0.0, 1.1],
                                [0.0, 0.0]]])

        function_output = normalized_cross_correlation(
            spiketrains, [-delay_time, delay_time]
        )

        assert np.allclose(function_output, test_output, 0.1)

    def test_total_spiking_probability_edges(self):
        files = ["SW/new_sim0_100.mat",
                 "BA/new_sim0_100.mat",
                 "CA/new_sim0_100.mat",
                 "ER05/new_sim0_100.mat",
                 "ER10/new_sim0_100.mat",
                 "ER15/new_sim0_100.mat",
                 ]

        for datafile in files:
            repo_base_path = 'unittest/functional_connectivity/' \
                             'total_spiking_probability_edges/data/'
            downloaded_dataset_path = download_datasets(repo_base_path +
                                                        datafile)

            spiketrains, original_data = load_spike_train_simulated(
                downloaded_dataset_path)

            connectivity_matrix, delay_matrix = \
                total_spiking_probability_edges(spiketrains)

            # Remove self-connections
            np.fill_diagonal(connectivity_matrix, 0)

            _, _, _, auc = roc_curve(connectivity_matrix, original_data)

            self.assertGreater(auc, 0.95)

# ====== HELPER FUNCTIONS ======


def classify_connections(connectivity_matrix: np.ndarray, threshold: int):
    connectivity_matrix_binarized = connectivity_matrix.copy()

    mask_excitatory = connectivity_matrix_binarized > threshold
    mask_inhibitory = connectivity_matrix_binarized < -threshold

    mask_left = ~ (mask_excitatory + mask_inhibitory)

    connectivity_matrix_binarized[mask_excitatory] = 1
    connectivity_matrix_binarized[mask_inhibitory] = -1
    connectivity_matrix_binarized[mask_left] = 0

    return connectivity_matrix_binarized


def confusion_matrix(estimate, original, threshold: int = 1):
    """
    Definition:
        - TP: Matches of connections are True Positive
        - FP: Mismatches are False Positive,
        - TN: Matches for non-existing synapses are True Negative
        - FN: mismatches are False Negative.
    """
    if not np.all(np.isin([-1, 0, 1], np.unique(estimate))):
        estimate = classify_connections(estimate, threshold)
    if not np.all(np.isin([-1, 0, 1], np.unique(original))):
        original = classify_connections(original, threshold)

    TP = (np.not_equal(estimate, 0) & np.not_equal(original, 0)).sum()

    TN = (np.equal(estimate, 0) & np.equal(original, 0)).sum()

    FP = (np.not_equal(estimate, 0) & np.equal(original, 0)).sum()

    FN = (np.equal(estimate, 0) & np.not_equal(original, 0)).sum()

    return TP, TN, FP, FN


def fall_out(TP: int, TN: int, FP: int, FN: int):
    FPR = FP / (FP + TN)
    return FPR


def sensitivity(TP: int, TN: int, FP: int, FN: int):
    TPR = TP / (TP + FN)
    return TPR


def roc_curve(estimate, original):
    tpr_list = []
    fpr_list = []

    max_threshold = max(np.max(np.abs(estimate)), 1)

    thresholds = np.linspace(max_threshold, 0, 30)

    for t in thresholds:
        conf_matrix = confusion_matrix(estimate, original, threshold=t)

        tpr_list.append(sensitivity(*conf_matrix))
        fpr_list.append(fall_out(*conf_matrix))

    auc = np.trapz(tpr_list, fpr_list)

    return tpr_list, fpr_list, thresholds, auc


def load_spike_train_simulated(path: Union[Path, str], bin_size=None,
                               t_stop=None,
                               ) -> Tuple[BinnedSpikeTrain, np.ndarray]:
    if isinstance(path, str):
        path = Path(path)

    if not bin_size:
        bin_size = 1 * ms

    data = loadmat(path, simplify_cells=True)["data"]

    if "asdf" not in data:
        raise ValueError('Incorrect Dataformat: Missing spiketrain_data in'
                         '"asdf"')

    spiketrain_data = data["asdf"]

    # Get number of electrodesa and recording_duration from last element of
    # data array
    n_electrodes, recording_duration_ms = spiketrain_data[-1]
    recording_duration_ms = recording_duration_ms * ms

    # Create spiketrains
    spiketrains = []
    for spiketrain_raw in spiketrain_data[0:n_electrodes]:
        spiketrains.append(
            SpikeTrain(
                spiketrain_raw * ms,
                t_stop=recording_duration_ms,
            )
        )

    spiketrains = BinnedSpikeTrain(spiketrains, bin_size=bin_size,
                                   t_stop=t_stop or recording_duration_ms)

    # Load original_data
    original_data = data['SWM'].T

    return spiketrains, original_data



# -*- coding: utf-8 -*-
"""
Identification of spectral properties in analog signals (e.g., the power
spectrum).

.. autosummary::
    :toctree: _toctree/spectral

    welch_psd
    welch_coherence
    multitaper_psd
    multitaper_cross_spectrum
    segmented_multitaper_cross_spectrum
    multitaper_coherence


:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import warnings

import neo
import numpy as np
import quantities as pq
import scipy.signal

__all__ = [
    "welch_psd",
    "welch_coherence",
    "multitaper_psd",
    "multitaper_cross_spectrum",
    "segmented_multitaper_cross_spectrum",
    "multitaper_coherence"
]


def welch_psd(signal, n_segments=8, len_segment=None,
              frequency_resolution=None, overlap=0.5, fs=1.0, window='hann',
              nfft=None, detrend='constant', return_onesided=True,
              scaling='density', axis=-1):
    """
    Estimates power spectrum density (PSD) of a given `neo.AnalogSignal`
    using Welch's method.

    The PSD is obtained through the following steps:

    1. Cut the given data into several overlapping segments. The degree of
       overlap can be specified by parameter `overlap` (default is 0.5,
       i.e. segments are overlapped by the half of their length).
       The number and the length of the segments are determined according
       to the parameters `n_segments`, `len_segment` or `frequency_resolution`.
       By default, the data is cut into 8 segments;

    2. Apply a window function to each segment. Hanning window is used by
       default. This can be changed by giving a window function or an
       array as parameter `window` (see Notes [2]);

    3. Compute the periodogram of each segment;

    4. Average the obtained periodograms to yield PSD estimate.

    Parameters
    ----------
    signal : neo.AnalogSignal or pq.Quantity or np.ndarray
        Time series data, of which PSD is estimated. When `signal` is
        `pq.Quantity` or `np.ndarray`, sampling frequency should be given
        through the keyword argument `fs`. Otherwise, the default value is
        used (`fs` = 1.0).
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_segment` or `frequency_resolution` is
        given.
        Default: 8
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it will be determined from other parameters.
        Default: None
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained PSD estimate in terms of
        the interval between adjacent frequency bins. When given as a `float`,
        it is taken as frequency in Hz.
        If None, it will be determined from other parameters.
        Default: None
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped)
    fs : pq.Quantity or float, optional
        Specifies the sampling frequency of the input time series. When the
        input is given as a `neo.AnalogSignal`, the sampling frequency is
        taken from its attribute and this parameter is ignored.
        Default: 1.0
    window : str or tuple or np.ndarray, optional
        Desired window to use.
        See Notes [2].
        Default: 'hann'
    nfft : int, optional
        Length of the FFT used.
        See Notes [2].
        Default: None
    detrend : str or function or False, optional
        Specifies how to detrend each segment.
        See Notes [2].
        Default: 'constant'
    return_onesided : bool, optional
        If True, return a one-sided spectrum for real data.
        If False return a two-sided spectrum.
        See Notes [2].
        Default: True
    scaling : {'density', 'spectrum'}, optional
        If 'density', computes the power spectral density where Pxx has units
        of V**2/Hz. If 'spectrum', computes the power spectrum where Pxx has
        units of V**2, if `signal` is measured in V and `fs` is measured in
        Hz.
        See Notes [2].
        Default: 'density'
    axis : int, optional
        Axis along which the periodogram is computed.
        See Notes [2].
        Default: last axis (-1)

    Returns
    -------
    freqs : pq.Quantity or np.ndarray
        Frequencies associated with the power estimates in `psd`.
        `freqs` is always a vector irrespective of the shape of the input
        data in `signal`.
        If `signal` is `neo.AnalogSignal` or `pq.Quantity`, a `pq.Quantity`
        array is returned.
        Otherwise, a `np.ndarray` containing frequency in Hz is returned.
    psd : pq.Quantity or np.ndarray
        PSD estimates of the time series in `signal`.
        If `signal` is `neo.AnalogSignal`, a `pq.Quantity` array is returned.
        Otherwise, the return is a `np.ndarray`.

    Raises
    ------
    ValueError
        If `overlap` is not in the interval `[0, 1)`.

        If `frequency_resolution` is not positive.

        If `frequency_resolution` is too high for the given data size.

        If `frequency_resolution` is None and `len_segment` is not a positive
        number.

        If `frequency_resolution` is None and `len_segment` is greater than the
        length of data at `axis`.

        If both `frequency_resolution` and `len_segment` are None and
        `n_segments` is not a positive number.

        If both `frequency_resolution` and `len_segment` are None and
        `n_segments` is greater than the length of data at `axis`.

    Notes
    -----
    1. The computation steps used in this function are implemented in
       `scipy.signal` module, and this function is a wrapper which provides
       a proper set of parameters to `scipy.signal.welch` function.
    2. The parameters `window`, `nfft`, `detrend`, `return_onesided`,
       `scaling`, and `axis` are directly passed to the `scipy.signal.welch`
       function. See the respective descriptions in the docstring of
       `scipy.signal.welch` for usage.
    3. When only `n_segments` is given, parameter `nperseg` of
       `scipy.signal.welch` function is determined according to the expression

       `signal.shape[axis] / (n_segments - overlap * (n_segments - 1))`

       converted to integer.

    See Also
    --------
    scipy.signal.welch
    welch_cohere

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.spectral import welch_psd
    >>> signal = neo.AnalogSignal(np.cos(np.linspace(0, 2 * np.pi, num=100)),
    ...     sampling_rate=20 * pq.Hz, units='mV')

    Sampling frequency will be taken as `signal.sampling_rate`.

    >>> freq, psd = welch_psd(signal)
    >>> freq
    array([ 0.        ,  0.90909091,  1.81818182,  2.72727273,  3.63636364,
            4.54545455,  5.45454545,  6.36363636,  7.27272727,  8.18181818,
            9.09090909, 10.        ]) * Hz

    >>> psd # noqa
    array([[1.09566410e-03, 2.33607943e-02, 1.35436832e-03, 6.74408723e-05,
            1.00810196e-05, 2.40079315e-06, 7.35821437e-07, 2.58361700e-07,
            9.44183422e-08, 3.14573483e-08, 6.82050475e-09, 1.18183354e-10]]) * mV**2/Hz


    """
    # 'hanning' window was removed with release of scipy 1.9.0, it was
    # deprecated since 1.1.0.
    if window == 'hanning':
        warnings.warn("'hanning' is deprecated and was removed from scipy "
                      "with release 1.9.0. Please use 'hann' instead",
                      DeprecationWarning)
        window = 'hann'
    # initialize a parameter dict (to be given to scipy.signal.welch()) with
    # the parameters directly passed on to scipy.signal.welch()
    params = {'window': window, 'nfft': nfft,
              'detrend': detrend, 'return_onesided': return_onesided,
              'scaling': scaling, 'axis': axis}

    # add the input data to params. When the input is AnalogSignal, the
    # data is added after rolling the axis for time index to the last
    data = np.asarray(signal)
    if isinstance(signal, neo.AnalogSignal):
        data = np.rollaxis(data, 0, len(data.shape))
    params['x'] = data

    # if the data is given as AnalogSignal, use its attribute to specify
    # the sampling frequency
    if hasattr(signal, 'sampling_rate'):
        params['fs'] = signal.sampling_rate.rescale('Hz').magnitude
    else:
        params['fs'] = fs

    if overlap < 0:
        raise ValueError("overlap must be greater than or equal to 0")
    elif 1 <= overlap:
        raise ValueError("overlap must be less then 1")

    # determine the length of segments (i.e. *nperseg*) according to given
    # parameters
    if frequency_resolution is not None:
        if frequency_resolution <= 0:
            raise ValueError("frequency_resolution must be positive")
        if isinstance(frequency_resolution, pq.quantity.Quantity):
            dF = frequency_resolution.rescale('Hz').magnitude
        else:
            dF = frequency_resolution
        nperseg = int(params['fs'] / dF)
        if nperseg > data.shape[axis]:
            raise ValueError("frequency_resolution is too high for the given "
                             "data size")
    elif len_segment is not None:
        if len_segment <= 0:
            raise ValueError("len_seg must be a positive number")
        elif data.shape[axis] < len_segment:
            raise ValueError("len_seg must be shorter than the data length")
        nperseg = len_segment
    else:
        if n_segments <= 0:
            raise ValueError("n_segments must be a positive number")
        elif data.shape[axis] < n_segments:
            raise ValueError("n_segments must be smaller than the data length")
        # when only *n_segments* is given, *nperseg* is determined by solving
        # the following equation:
        #  n_segments * nperseg - (n_segments-1) * overlap * nperseg =
        #     data.shape[-1]
        #  --------------------   ===============================  ^^^^^^^^^^^
        # summed segment lengths        total overlap              data length
        nperseg = int(data.shape[axis] / (n_segments - overlap * (
            n_segments - 1)))
    params['nperseg'] = nperseg
    params['noverlap'] = int(nperseg * overlap)

    freqs, psd = scipy.signal.welch(**params)

    # attach proper units to return values
    if isinstance(signal, pq.quantity.Quantity):
        if 'scaling' in params and params['scaling'] == 'spectrum':
            psd = psd * signal.units * signal.units
        else:
            psd = psd * signal.units * signal.units / pq.Hz
        freqs = freqs * pq.Hz

    return freqs, psd


def multitaper_psd(signal, fs=1, nw=4, num_tapers=None, peak_resolution=None,
                   attach_units=True):
    """
    Estimates power spectrum density (PSD) of a given 'neo.AnalogSignal'
    using the Multitaper method.

    The PSD is obtained through the following steps:

    1. Calculate 'num_tapers' approximately independent estimates of the
       spectrum by multiplying the signal with the discrete prolate spheroidal
       functions (also known as Slepian function) and calculate the PSD of each
       tapered signal.

    2. Average the approximately independent estimates of each signal to
       decrease overall variance of the estimates

    Parameters
    ----------
    signal : neo.AnalogSignal or pq.Quantity or np.ndarray
        Time series data of which PSD is estimated. When `signal` is np.ndarray
        sampling frequency should be given through keyword argument `fs`.
        Signal should be passed as (n_channels, n_samples)
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0.
    nw : float, optional
        Time bandwidth product
        Default: 4.0.
    num_tapers : int, optional
        Number of tapers used in 1. to obtain estimate of PSD. By default,
        [2*nw] - 1 is chosen.
        Default: None.
    peak_resolution : pq.Quantity float, optional
        Quantity in Hz determining the number of tapers used for analysis.
        Fine peak resolution --> low numerical value --> low number of tapers
        High peak resolution --> high numerical value --> high number of tapers
        When given as a `float`, it is taken as frequency in Hz.
        Default: None.
    attach_units: bool, optional
        If True and signals is an instance of pq.Quantity, units are attached
        to the estimated cross spectrum.
        Default: True

    Notes
    -----
    1. There is a parameter hierarchy regarding nw, num_tapers and
       peak_resolution. If peak_resolution is provided, it determines both nw
       and the num_tapers. Specifying num_tapers has an effect only if
       peak_resolution is not provided.

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with power estimate in `psd`
    psd : np.ndarray
        PSD estimate of the time series in `signal`

    Raises
    ------
    ValueError
        If `peak_resolution` is not a positive number.

        If `peak_resolution` is None and `num_tapers` is not a positive number.

    TypeError
        If `peak_resolution` is None and `num_tapers` is not an int.
    """

    # When the input is AnalogSignal, the data is added after rolling the axis
    # for time index to the last
    data = np.asarray(signal)
    if isinstance(signal, neo.AnalogSignal):
        data = np.moveaxis(data, 0, 1)
        fs = signal.sampling_rate

    if isinstance(fs, pq.Quantity):
        fs = fs.rescale('Hz').magnitude

    # Add a dim if data has only one dimension
    if data.ndim == 1:
        data = data[np.newaxis, :]

    length_signal = np.shape(data)[1]

    # If peak resolution is pq.Quantity, get magnitude
    if isinstance(peak_resolution, pq.quantity.Quantity):
        peak_resolution = peak_resolution.rescale('Hz').magnitude

    # Determine time-halfbandwidth product from given parameters
    if peak_resolution is not None:
        if peak_resolution <= 0:
            raise ValueError("peak_resolution must be positive")
        nw = length_signal / fs * peak_resolution / 2
        num_tapers = int(np.floor(2*nw) - 1)

    if num_tapers is None:
        num_tapers = int(np.floor(2*nw) - 1)
    else:
        if not isinstance(num_tapers, int):
            raise TypeError("num_tapers must be integer")
        if num_tapers <= 0:
            raise ValueError("num_tapers must be positive")

    # Generate frequencies of PSD estimate
    freqs = np.fft.rfftfreq(length_signal, d=1/fs)

    # Generate Slepian sequences
    slepian_fcts = scipy.signal.windows.dpss(M=length_signal,
                                             NW=nw,
                                             Kmax=num_tapers,
                                             sym=False)

    # Calculate approximately independent spectrum estimates
    # Use broadcasting to match dim for point-wise multiplication
    # Shape: (n_channels, n_tapers, n_samples)
    tapered_signal = data[:, np.newaxis, :] * slepian_fcts

    # Determine Fourier transform of tapered signal
    spectrum_estimates = np.abs(np.fft.rfft(tapered_signal, axis=-1))**2
    spectrum_estimates[..., 1:] *= 2

    # Average Fourier transform windowed signal
    psd = np.mean(spectrum_estimates, axis=-2) / fs

    # Attach proper units to return values
    if isinstance(signal, pq.quantity.Quantity) and attach_units:
        psd = psd * signal.units * signal.units / pq.Hz
        freqs = freqs * pq.Hz

    return freqs, psd


def segmented_multitaper_psd(signal, n_segments=1, len_segment=None,
                             frequency_resolution=None, overlap=0.5, fs=1,
                             nw=4, num_tapers=None, peak_resolution=None):
    """
    Estimates power spectrum density (PSD) of a given 'neo.AnalogSignal'
    using Multitaper method

    The PSD is obtained through the following steps:

    1. Cut the given data into several overlapping segments. The degree of
       overlap can be specified by parameter `overlap` (default is 0.5,
       i.e. segments are overlapped by the half of their length).
       The number and the length of the segments are determined according
       to the parameters `n_segments`, `len_segment` or `frequency_resolution`.
       By default, the data is not cut into  segments;

    2. Calculate 'num_tapers' approximately independent estimates of the
       spectrum by multiplying the signal with the discrete prolate spheroidal
       functions (also known as Slepian function) and calculate the PSD of each
       tapered segment

    3. Average the approximately independent estimates of each segment to
       decrease overall variance of the estimates

    4. Average the obtained estimates for each segment

    Parameters
    ----------
    signal : neo.AnalogSignal or pq.Quantity or np.ndarray
        Time series data of which PSD is estimated. When `signal` is np.ndarray
        sampling frequency should be given through keyword argument `fs`.
        Signal should be passed as (n_channels, n_samples)
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0.
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_segment` or `frequency_resolution` is
        given.
        Default: 1.
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it will be determined from other parameters.
        Default: None.
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained PSD estimate in terms of
        the interval between adjacent frequency bins. When given as a `float`,
        it is taken as frequency in Hz.
        If None, it will be determined from other parameters.
        Default: None.
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped).
    nw : float, optional
        Time bandwidth product
        Default: 4.0.
    num_tapers : int, optional
        Number of tapers used in 1. to obtain estimate of PSD. By default,
        [2*nw] - 1 is chosen.
        Default: None.
    peak_resolution : pq.Quantity float, optional
        Quantity in Hz determining the number of tapers used for analysis.
        Fine peak resolution --> low numerical value --> low number of tapers
        High peak resolution --> high numerical value --> high number of tapers
        When given as a `float`, it is taken as frequency in Hz.
        Default: None.

    Notes
    -----
    1. There is a parameter hierarchy regarding n_segments and len_segment. The
       former parameter is ignored if the latter one is passed.

    2. There is a parameter hierarchy regarding nw, num_tapers and
       peak_resolution. If peak_resolution is provided, it determines both nw
       and the num_tapers. Specifying num_tapers has an effect only if
       peak_resolution is not provided.

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with power estimate in `psd`
    psd : np.ndarray
        PSD estimate of the time series in `signal`

    Raises
    ------
    ValueError
        If `peak_resolution` is None and `num_tapers` is not a positive number.

        If `frequency_resolution` is too high for the given data size.

        If `frequency_resolution` is None and `len_segment` is not a positive
        number.

        If `frequency_resolution` is None and `len_segment` is greater than the
        length of data at `axis`.

        If both `frequency_resolution` and `len_segment` are None and
        `n_segments` is not a positive number.

        If both `frequency_resolution` and `len_segment` are None and
        `n_segments` is greater than the length of data at `axis`.

    TypeError
        If `peak_resolution` is None and `num_tapers` is not an int.
    """

    # When the input is AnalogSignal, the data is added after rolling the axis
    # for time index to the last
    data = np.asarray(signal)
    if isinstance(signal, neo.AnalogSignal):
        data = np.moveaxis(data, 0, 1)
        fs = signal.sampling_rate

    # Add a dim if data has only one dimension
    if data.ndim == 1:
        data = data[np.newaxis, :]

    # Number of data points in time series
    length_signal = np.shape(data)[1]

    psd_params_dict = {
        'nw': nw,
        'num_tapers': num_tapers,
        'peak_resolution': peak_resolution,
        'attach_units': False
    }

    freqs, psd = _segmented_apply_func(
        data=data, fs=fs, n_segments=n_segments, len_segment=len_segment,
        overlap=overlap, frequency_resolution=frequency_resolution,
        func=multitaper_psd, func_params_dict=psd_params_dict)

    # Attach proper units to return values
    if isinstance(signal, pq.quantity.Quantity):
        psd = psd * signal.units * signal.units / pq.Hz
        freqs = freqs * pq.Hz

    return freqs, psd


def multitaper_cross_spectrum(signals, fs=1.0, nw=4.0, num_tapers=None,
                              peak_resolution=None, return_onesided=True,
                              attach_units=True):
    """
    Estimates the cross spectrum of a given `neo.AnalogSignal` using the
    Multitaper method.

    The cross spectrum is obtained through the following steps:

    1. Obtain approximately independent estimates of the spectrum for each
       signal by multiplying it with tapering functions (obtained as
       the discrete prolate spheroidal functions, also known as slepian
       functions) and calculating the cross spectrum for the tapered signals.
       The number of tapering functions (and hence the number of the estimates)
       is specified by the parameter `num_tapers`.

    2. Average the approximately independent estimates of each tapered
       instance of the signal to decrease overall variance of the estimates.

    Parameters
    ----------
    signals : neo.AnalogSignal or pq.Quantity or np.ndarray
        Time series data of which PSD is estimated. When `signal` is
        `np.ndarray` or `pq.Quantity`, it needs to be passed as a 2-dimensional
        array of shape (n_channels, n_samples), and the data sampling frequency
        should be given through the keyword argument `fs`.
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0
    nw : float, optional
        Time-halfbandwidth product. This parameter can be used to determine the
        number of tapers following the equation:

            num_tapers = 2*nw - 1

        It can be determined by multiplying the duration of the signal with the
        desired half-peak resolution frequency:

            n_samples/fs * peak_resolution/2.

        Default: 4.0
    num_tapers : int, optional
        Number of tapers used in step 2 (see above) to obtain estimate of PSD.
        By default, [2*nw] - 1 is chosen.
        Default: None (see Notes [1])
    peak_resolution : pq.Quantity float, optional
        Quantity in Hz determining the number of tapers used for analysis.
        Fine peak resolution --> low numerical value --> low number of tapers
        High peak resolution --> high numerical value --> high number of tapers
        When given as a `float`, it is taken as frequency in Hz.
        Default: None (see Notes [1])
    return_onesided : bool, optional
        If True, return a one-sided spectrum for real data.
        If False return a two-sided spectrum.
        Default: True
    attach_units : bool, optional
        If True and signals is instance of pq.Quantity, units are attached to
        the estimated cross spectrum.
        Default: True

    Notes
    -----
    1. There is a parameter hierarchy regarding `nw`, `num_tapers` and
       `peak_resolution`. If peak_resolution is provided, it determines both
       `nw` and the `num_tapers`. Specifying `num_tapers` has an effect only if
       `peak_resolution` is not provided.

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with cross spectrum estimate in `cross_spec`. If
        signals is an instance of `neo.AnalogSignal` or `Quantity` and the
        `attach_units` is set to `True`, the returned values will have units of
        Hz.
    cross_spec : np.ndarray
        Estimate of the cross spectrum of time series in `signals`. If signals
        is an instance of `neo.AnalogSignal` or `Quantity` and the
        `attach_units` is set to `True`, the returned values will have units
        attached to them (i.e. signals.units ** 2/Hz).

    Raises
    ------
    ValueError
        If `peak_resolution` is None and `num_tapers` is not a positive number.

    TypeError
        If `peak_resolution` is None and `num_tapers` is not an int.
    """
    # When the input is AnalogSignal, fetch the underlying numpy array and swap
    # axes from (n_samples, n_channels) to (n_channels, n_samples)
    data = np.asarray(signals)
    if isinstance(signals, neo.AnalogSignal):
        data = np.moveaxis(data, 0, 1)

    # Number of data points in time series
    length_signal = np.shape(data)[1]

    # If the data is given as AnalogSignal, use its attribute to specify the
    # sampling frequency
    if hasattr(signals, 'sampling_rate'):
        fs = signals.sampling_rate.rescale('Hz')

    # If fs and peak resolution is pq.Quantity, get magnitude
    if isinstance(fs, pq.quantity.Quantity):
        fs = fs.rescale('Hz').magnitude

    # Determine time-halfbandwidth product from given parameters
    if peak_resolution is not None:
        if isinstance(peak_resolution, pq.quantity.Quantity):
            peak_resolution = peak_resolution.rescale('Hz').magnitude
        if peak_resolution <= 0:
            raise ValueError("peak_resolution must be positive")
        nw = length_signal / fs * peak_resolution / 2
        num_tapers = int(np.floor(2*nw) - 1)

    if num_tapers is None:
        num_tapers = int(np.floor(2*nw) - 1)
    else:
        if not isinstance(num_tapers, int):
            raise TypeError("num_tapers must be integer")
        if num_tapers <= 0:
            raise ValueError("num_tapers must be positive")

    # Get slepian functions (sym='False' used for spectral analysis)
    slepian_fcts = scipy.signal.windows.dpss(M=length_signal,
                                             NW=nw,
                                             Kmax=num_tapers,
                                             sym=False)

    # Use broadcasting to match dime for point-wise multiplication
    tapered_signals = (data[:, np.newaxis] * slepian_fcts)

    if return_onesided:
        # Determine frequencies for real Fourier transform
        freqs = np.fft.rfftfreq(length_signal, 1 / fs)
        # Determine real Fourier transform of tapered signal
        spectrum_estimates = np.fft.rfft(tapered_signals, axis=-1)

    else:
        # Determine frequencies for full Fourier transform
        freqs = np.fft.fftfreq(length_signal, 1 / fs)
        # Determine full Fourier transform of tapered signal
        spectrum_estimates = np.fft.fft(tapered_signals, axis=-1)

    temp = (spectrum_estimates[np.newaxis, :, :, :]
            * np.conjugate(spectrum_estimates[:, np.newaxis, :, :]))

    # Average Fourier transform windowed signal
    cross_spec = np.mean(temp, axis=-2, dtype=np.complex64) / fs

    if attach_units and isinstance(signals, pq.quantity.Quantity):
        freqs = freqs * pq.Hz
        cross_spec = cross_spec * signals.units * signals.units / pq.Hz

    return freqs, cross_spec


def _segmented_apply_func(data, func, fs=1.0, n_segments=1, len_segment=None,
                          frequency_resolution=None, overlap=0.5,
                          func_params_dict=None):
    """
    Estimate a spectral measure of a signal by applying it to segments of a
    signal and then averaging over the segments.
    The spectral measure is computed on each segment by a given function.
    The underlying assumption is the spectral measures computed for the
    segments are approximately independent.

    The spectral measure is obtained through the following steps

    1. Cut the given data into several overlapping segments. The degree of
       overlap can be specified by parameter `overlap` (default is 0.5,
       i.e. segments are overlapped by the half of their length).
       The number and the length of the segments are determined according
       to the parameters `n_segments`, `len_segment` or `frequency_resolution`.
       By default, the `data` is not cut into segments;

    2. Apply the function calculating the desired spectral measure.

    3. Average the obtained estimates for each segment

    Parameters
    ----------
    data : np.ndarray
        Time series data of which spectral measure is estimated. `data`
        is `np.ndarray` or `pq.Quantity` of shape (n_channels, n_samples), and
        the data sampling frequency should be given through the keyword
        argument `fs` in `func_params_dict`.
    func : callable
        Function calculating spectral measure.
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_segment` or `frequency_resolution` is
        given.
        Default: 1
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it will be determined from other parameters.
        Default: None
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained PSD estimate in terms of
        the interval between adjacent frequency bins. When given as a `float`,
        it is taken as frequency in Hz.
        If None, it will be determined from other parameters.
        Default: None
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped)
    func_params_dict : dict, optional
        Arguments for function `func` returning spectral measure.

    Notes
    -----
    1. There is a parameter hierarchy regarding `frequency_resolution`,
    `len_segment` and `n_segments` (in decreasing order of priority). If the
    former parameter is passed, the latter parameter(s) is/are ignored.

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with the estimated spectral measure
    avg_estimate : np.ndarray
        Estimated spectral measure for segmented data

    Raises
    ------
    ValueError
        If `frequency_resolution` is too high for the given data size.

        If `frequency_resolution` is None and `len_segment` is not a positive
        number.

        If `frequency_resolution` is None and `len_segment` is greater than the
        length of data at `axis`.

        If both `frequency_resolution` and `len_segment` are None and
        `n_segments` is not a positive number.

        If both `frequency_resolution` and `len_segment` are None and
        `n_segments` is greater than the length of data at `axis`.
    """
    # Number of data points in time series
    length_signal = np.shape(data)[1]

    # It is assumed that data has shape (n_channels, n_samples). To fetch
    # samples set correct axis
    axis = -1

    func_params_dict.update(dict(fs=fs))

    # If fs and peak resolution is pq.Quantity, get magnitude
    if isinstance(fs, pq.quantity.Quantity):
        fs = fs.rescale('Hz').magnitude

    # Determine length per segment - n_per_seg
    if frequency_resolution is not None:
        if frequency_resolution <= 0:
            raise ValueError("frequency_resolution must be positive")
        if isinstance(frequency_resolution, pq.quantity.Quantity):
            dF = frequency_resolution.rescale('Hz').magnitude
        else:
            dF = frequency_resolution
        n_per_seg = int(fs / dF)
        if n_per_seg > data.shape[axis]:
            raise ValueError("frequency_resolution is too high for the given "
                             "data size")
    elif len_segment is not None:
        if len_segment <= 0:
            raise ValueError("len_seg must be a positive number")
        elif data.shape[axis] < len_segment:
            raise ValueError("len_seg must be shorter than the data length")
        n_per_seg = len_segment
    else:
        if n_segments <= 0:
            raise ValueError("n_segments must be a positive number")
        elif data.shape[axis] < n_segments:
            raise ValueError("n_segments must be smaller than the data length")
        # when only *n_segments* is given, *n_per_seg* is determined by solving
        # the following equation:
        #  n_segments * n_per_seg - (n_segments-1) * overlap * n_per_seg =
        #     data.shape[-1]
        #  --------------------   ===============================  ^^^^^^^^^^^
        # summed segment lengths        total overlap              data length
        n_per_seg = int(data.shape[axis] /
                        (n_segments - overlap * (n_segments - 1)))

    n_overlap = int(n_per_seg * overlap)
    n_overlap_step = n_per_seg - n_overlap
    n_segments = int((length_signal - n_overlap) / (n_per_seg - n_overlap))

    # Generate frequencies for spectral measure estimate
    if func_params_dict.get('return_onesided'):
        freqs = np.fft.rfftfreq(n_per_seg, d=1/fs)
    elif func_params_dict.get('return_onesided') is None:
        # multitaper_psd uses rfft (i.e. no return_onesided parameter)
        freqs = np.fft.rfftfreq(n_per_seg, d=1/fs)
    else:
        freqs = np.fft.fftfreq(n_per_seg, d=1/fs)

    # Zero-pad signal to fit segment length
    remainder = length_signal % n_overlap_step

    data = np.pad(data, [(0, 0), (0, remainder)],
                  mode='constant', constant_values=0)

    # Generate array for storing cross spectra estimates of segments
    seg_estimates = np.zeros((n_segments,
                              data.shape[0],
                              data.shape[0],
                              len(freqs)),
                             dtype=np.complex64)

    n_overlap_step = n_per_seg - n_overlap

    for i in range(n_segments):

        _, estimate = func(
            data[:, i * n_overlap_step:i * n_overlap_step + n_per_seg],
            **func_params_dict)

        # Workaround for mismatched dimensions
        if estimate.ndim != seg_estimates.ndim - 1:  # Multitaper PSD
            seg_estimates[i] = estimate[:, np.newaxis, :]
        else:
            seg_estimates[i] = estimate

    avg_estimate = np.mean(seg_estimates, axis=0)

    return freqs, avg_estimate


def segmented_multitaper_cross_spectrum(signals, n_segments=1,
                                        len_segment=None,
                                        frequency_resolution=None, overlap=0.5,
                                        fs=1.0, nw=4.0, num_tapers=None,
                                        peak_resolution=None,
                                        return_onesided=True):
    """
    Estimates the cross spectrum of a given `neo.AnalogSignal` using the
    Multitaper method on segments of the data.

    The segmented cross spectrum is obtained through the following steps:

    1. Cut the given data into several overlapping segments. The degree of
       overlap can be specified by parameter `overlap` (default is 0.5,
       i.e. segments are overlapped by the half of their length).
       The number and the length of the segments are determined according
       to the parameters `n_segments`, `len_segment` or `frequency_resolution`.
       By default, the `data` is not cut into  segments.

    2. Calculate 'num_tapers' approximately independent estimates of the
       spectrum by multiplying the signal with the discrete prolate spheroidal
       functions (also known as Slepian function) and calculate the PSD of each
       tapered segment.

    3. Average the approximately independent estimates of each segment to
       decrease overall variance of the estimates.

    4. Average the obtained estimates for each segment.

    The estimation is done by applying `_segmented_apply_func` to
    :func:`multitaper_cross_spectrum`. See the latter function for further
    documentation.

    Parameters
    ----------
    signals : neo.AnalogSignal or pq.Quantity or np.ndarray
        Time series data of which CSD is estimated. When `signal` is
        `np.ndarray` or `pq.Quantity`, it needs to be passed as a 2-dimensional
        array of shape (n_channels, n_samples), and the data sampling frequency
        should be given through the keyword argument `fs`.
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_segment` or `frequency_resolution` is
        given.
        Default: 1
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it will be determined from other parameters.
        Default: None
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained PSD estimate in terms of
        the interval between adjacent frequency bins. When given as a `float`,
        it is taken as frequency in Hz.
        If None, it will be determined from other parameters.
        Default: None
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped)
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0
    nw : float, optional
        Time-halfbandwidth product. This parameter can be used to determine the
        number of tapers following the equation:

            num_tapers = 2*nw - 1

        It can be determined by multiplying the duration of the signal with the
        desired half-peak resolution frequency:

            n_samples/fs * peak_resolution/2.

        Default: 4.0
    num_tapers : int, optional
        Number of tapers used in step 2 (see above) to obtain estimate of PSD.
        By default, [2*nw] - 1 is chosen.
        Default: None
    peak_resolution : pq.Quantity float, optional
        Quantity in Hz determining the number of tapers used for analysis.
        Fine peak resolution --> low numerical value --> low number of tapers
        High peak resolution --> high numerical value --> high number of tapers
        When given as a `float`, it is taken as frequency in Hz.
        Default: None
    return_onesided : bool, optional
        If True, return a one-sided spectrum for real data.
        If False return a two-sided spectrum.
        Default: True

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with the cross spectrum estimate in
        `cross_spec`. If signals is an instance of `neo.AnalogSignal` or
        `Quantity`, the returned values will have units of Hz.
    cross_spec : np.ndarray
        Estimate of the cross spectrum of time series in `signals`. If signals
        is an instance of `neo.AnalogSignal` or `Quantity`, the returned
        values will have units attached to them (i.e. signals.units ** 2/Hz).

    See Also
    --------
    :func:`multitaper_cross_spectrum`


    """
    # When the input is AnalogSignal, fetch the underlying numpy array and swap
    # axes from (n_samples, n_channels) to (n_channels, n_samples)
    data = np.asarray(signals)
    if isinstance(signals, neo.AnalogSignal):
        data = np.moveaxis(data, 0, 1)
        fs = signals.sampling_rate

    # Initialize argument dictionary for multitaper_cross_spectrum called on
    # segments
    cross_spec_params_dict = {
        'nw': nw,
        'num_tapers': num_tapers,
        'peak_resolution': peak_resolution,
        'return_onesided': return_onesided,
        'attach_units': False}

    # Apply segmentation
    freqs, cross_spec = _segmented_apply_func(
        data=data, fs=fs, n_segments=n_segments, len_segment=len_segment,
        overlap=overlap, frequency_resolution=frequency_resolution,
        func=multitaper_cross_spectrum,
        func_params_dict=cross_spec_params_dict)

    # Attach proper units to return values
    if isinstance(signals, pq.quantity.Quantity):
        cross_spec = cross_spec * signals.units * signals.units / pq.Hz
        freqs = freqs * pq.Hz

    return freqs, cross_spec


def multitaper_coherence(signal_i, signal_j, n_segments=1, len_segment=None,
                         frequency_resolution=None, overlap=0.5, fs=1,
                         nw=4, num_tapers=None, peak_resolution=None):
    """
    Estimates the magnitude-squared coherence and phase-lag of two given
    `neo.AnalogSignal` using the Multitaper method.
    The function calls `segmented_multitaper_cross_spectrum` internally and
    thus both functions share parts of the respective signatures.

    .. math::
        C(\omega)=\\frac{|S_{xy}(\omega)|^2}{S_{xx}(\omega)S_{yy}(\omega)}

    Parameters
    ----------
    signal_i : neo.AnalogSignal or pq.Quantity or np.ndarray
        First time series data of the pair between which coherence is
        computed.
    signal_j : neo.AnalogSignal or pq.Quantity or np.ndarray
        Second time series data of the pair between which coherence is
        computed.
        The shapes and the sampling frequencies of `signal_i` and `signal_j`
        must be identical. When `signal_i` and `signal_j` are not
        `neo.AnalogSignal`, sampling frequency should be specified through the
        keyword argument `fs`. Otherwise, the default value is used
        (`fs` = 1.0).
    fs : float, optional
        Specifies the sampling frequency of the input time series
        Default: 1.0
    nw : float, optional
        Time bandwidth product
        Default: 4.0
    num_tapers : int, optional
        Number of tapers used in 1. to obtain estimate of PSD. By default,
        [2*nw] - 1 is chosen.
        Default: None
    peak_resolution : pq.Quantity float, optional
        Quantity in Hz determining the number of tapers used for analysis.
        Fine peak resolution --> low numerical value --> low number of tapers
        High peak resolution --> high numerical value --> high number of tapers
        When given as a `float`, it is taken as frequency in Hz.
        Default: None.
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_segment` or `frequency_resolution` is
        given.
        Default: 1
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it will be determined from other parameters.
        Default: None
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained coherence estimate in
        terms of the interval between adjacent frequency bins. When given as a
        `float`, it is taken as frequency in Hz.
        If None, it will be determined from other parameters.
        Default: None
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped)

    Returns
    -------
    freqs : np.ndarray
        Frequencies associated with the magnitude-squared coherence estimate
    coherence : np.ndarray
        Magnitude-squared coherence estimate
    phase_lag : np.ndarray
        Phase lags associated with the magnitude-square coherence estimate

    """
    if isinstance(signal_i, neo.core.AnalogSignal) and \
            isinstance(signal_j, neo.core.AnalogSignal):
        signals = signal_i.merge(signal_j)
    elif isinstance(signal_i, np.ndarray) and isinstance(signal_j, np.ndarray):
        signals = np.vstack([signal_i, signal_j])

    freqs, Pxy = segmented_multitaper_cross_spectrum(
        signals=signals, n_segments=n_segments, len_segment=len_segment,
        frequency_resolution=frequency_resolution, overlap=overlap, fs=fs,
        nw=nw, num_tapers=num_tapers, peak_resolution=peak_resolution)

    # Calculate magnitude-squared coherence.
    coherence = np.abs(Pxy[0, 1]) ** 2 / (Pxy[0, 0].real * Pxy[1, 1].real)

    phase_lag = np.angle(Pxy[0, 1])

    return freqs, coherence, phase_lag


def welch_coherence(signal_i, signal_j, n_segments=8, len_segment=None,
                    frequency_resolution=None, overlap=0.5, fs=1.0,
                    window='hann', nfft=None, detrend='constant',
                    scaling='density', axis=-1):
    r"""
    Estimates coherence between a given pair of analog signals.

    The estimation is performed with Welch's method: the given pair of data
    are cut into short segments, cross-spectra are calculated for each pair of
    segments, and the cross-spectra are averaged and normalized by respective
    auto-spectra.

    By default, the data are cut into 8 segments with 50% overlap between
    neighboring segments. These numbers can be changed through respective
    parameters.

    Parameters
    ----------
    signal_i : neo.AnalogSignal or pq.Quantity or np.ndarray
        First time series data of the pair between which coherence is
        computed.
    signal_j : neo.AnalogSignal or pq.Quantity or np.ndarray
        Second time series data of the pair between which coherence is
        computed.
        The shapes and the sampling frequencies of `signal_i` and `signal_j`
        must be identical. When `signal_i` and `signal_j` are not
        `neo.AnalogSignal`, sampling frequency should be specified through the
        keyword argument `fs`. Otherwise, the default value is used
        (`fs` = 1.0).
    n_segments : int, optional
        Number of segments. The length of segments is adjusted so that
        overlapping segments cover the entire stretch of the given data. This
        parameter is ignored if `len_seg` or `frequency_resolution` is given.
        Default: 8
    len_segment : int, optional
        Length of segments. This parameter is ignored if `frequency_resolution`
        is given. If None, it is determined from other parameters.
        Default: None
    frequency_resolution : pq.Quantity or float, optional
        Desired frequency resolution of the obtained coherence estimate in
        terms of the interval between adjacent frequency bins. When given as a
        `float`, it is taken as frequency in Hz.
        If None, it is determined from other parameters.
        Default: None
    overlap : float, optional
        Overlap between segments represented as a float number between 0 (no
        overlap) and 1 (complete overlap).
        Default: 0.5 (half-overlapped)
    fs : pq.Quantity or float, optional
        Specifies the sampling frequency of the input time series. When the
        input time series are given as `neo.AnalogSignal`, the sampling
        frequency is taken from their attribute and this parameter is ignored.
        Default: 1.0
    window : str or tuple or np.ndarray, optional
        Desired window to use.
        See Notes [1].
        Default: 'hann'
    nfft : int, optional
        Length of the FFT used.
        See Notes [1].
        Default: None
    detrend : str or function or False, optional
        Specifies how to detrend each segment.
        See Notes [1].
        Default: 'constant'
    scaling : {'density', 'spectrum'}, optional
        If 'density', computes the power spectral density where Pxx has units
        of V**2/Hz. If 'spectrum', computes the power spectrum where Pxx has
        units of V**2, if `signal` is measured in V and `fs` is measured in
        Hz.
        See Notes [1].
        Default: 'density'
    axis : int, optional
        Axis along which the periodogram is computed.
        See Notes [1].
        Default: last axis (-1)

    Returns
    -------
    freqs : pq.Quantity or np.ndarray
        Frequencies associated with the estimates of coherency and phase lag.
        `freqs` is always a vector irrespective of the shape of the input
        data. If `signal_i` and `signal_j` are `neo.AnalogSignal` or
        `pq.Quantity`, a `pq.Quantity` array is returned. Otherwise, a
        `np.ndarray` containing frequency in Hz is returned.
    coherency : np.ndarray
        Estimate of coherency between the input time series. For each
        frequency, coherency takes a value between 0 and 1, with 0 or 1
        representing no or perfect coherence, respectively.
        When the input arrays `signal_i` and `signal_j` are multidimensional,
        `coherency` is of the same shape as the inputs, and the frequency is
        indexed depending on the type of the input. If the input is
        `neo.AnalogSignal`, the first axis indexes frequency. Otherwise,
        frequency is indexed by the last axis.
    phase_lag : pq.Quantity or np.ndarray
        Estimate of phase lag in radian between the input time series. For
        each frequency, phase lag takes a value between :math:`-\pi` and
        :math:`\pi`, with positive values meaning phase precession of
        `signal_i` ahead of `signal_j`, and vice versa. If `signal_i` and
        `signal_j` are `neo.AnalogSignal` or `pq.Quantity`, a `pq.Quantity`
        array is returned. Otherwise, a `np.ndarray` containing phase lag in
        radian is returned. The axis for frequency index is determined in the
        same way as for `coherency`.

    Raises
    ------
    ValueError
        Same as in :func:`welch_psd`.

    Notes
    -----
    1. The computation steps used in this function are implemented in
       `scipy.signal` module, and this function is a wrapper which provides
       a proper set of parameters to `scipy.signal.welch` function.
    2. The parameters `window`, `nfft`, `detrend`, `return_onesided`,
       `scaling`, and `axis` are directly passed to the `scipy.signal.welch`
       function. See the respective descriptions in the docstring of
       `scipy.signal.welch` for usage.
    3. When only `n_segments` is given, parameter `nperseg` of
       `scipy.signal.welch` function is determined according to the expression

       `signal.shape[axis] / (n_segments - overlap * (n_segments - 1))`

       converted to integer.

    See Also
    --------
    welch_psd

    Examples
    --------
    >>> import neo
    >>> import numpy as np
    >>> import quantities as pq
    >>> from elephant.spectral import welch_coherence
    >>> signal = neo.AnalogSignal(np.cos(np.linspace(0, 2 * np.pi, num=100)),
    ...     sampling_rate=20 * pq.Hz, units='mV')

    Sampling frequency will be taken as `signal.sampling_rate`.

    >>> freq, coherency, phase_lag = welch_coherence(signal, signal)
    >>> freq
    array([ 0.        ,  0.90909091,  1.81818182,  2.72727273,  3.63636364,
            4.54545455,  5.45454545,  6.36363636,  7.27272727,  8.18181818,
            9.09090909, 10.        ]) * Hz


    >>> coherency.flatten()
    array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])
    >>> phase_lag.flatten() # doctest: +SKIP
    array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]) * rad

    """

    # TODO: code duplication with welch_psd()
    # 'hanning' window was removed with release of scipy 1.9.0, it was
    # deprecated since 1.1.0.
    if window == 'hanning':
        warnings.warn("'hanning' is deprecated and was removed from scipy "
                      "with release 1.9.0. Please use 'hann' instead",
                      DeprecationWarning)
        window = 'hann'

    # initialize a parameter dict for scipy.signal.csd()
    params = {'window': window, 'nfft': nfft,
              'detrend': detrend, 'scaling': scaling, 'axis': axis}

    # When the input is AnalogSignal, the axis for time index is rolled to
    # the last
    xdata = np.asarray(signal_i)
    ydata = np.asarray(signal_j)
    if isinstance(signal_i, neo.AnalogSignal):
        xdata = np.rollaxis(xdata, 0, len(xdata.shape))
        ydata = np.rollaxis(ydata, 0, len(ydata.shape))

    # if the data is given as AnalogSignal, use its attribute to specify
    # the sampling frequency
    if hasattr(signal_i, 'sampling_rate'):
        params['fs'] = signal_i.sampling_rate.rescale('Hz').magnitude
    else:
        params['fs'] = fs

    if overlap < 0:
        raise ValueError("overlap must be greater than or equal to 0")
    elif 1 <= overlap:
        raise ValueError("overlap must be less then 1")

    # determine the length of segments (i.e. *nperseg*) according to given
    # parameters
    if frequency_resolution is not None:
        if isinstance(frequency_resolution, pq.quantity.Quantity):
            dF = frequency_resolution.rescale('Hz').magnitude
        else:
            dF = frequency_resolution
        nperseg = int(params['fs'] / dF)
        if nperseg > xdata.shape[axis]:
            raise ValueError("frequency_resolution is too high for the given"
                             "data size")
    elif len_segment is not None:
        if len_segment <= 0:
            raise ValueError("len_seg must be a positive number")
        if xdata.shape[axis] < len_segment:
            raise ValueError("len_seg must be shorter than the data length")
        nperseg = len_segment
    else:
        if n_segments <= 0:
            raise ValueError("n_segments must be a positive number")
        if xdata.shape[axis] < n_segments:
            raise ValueError("n_segments must be smaller than the data length")
        # when only *n_segments* is given, *nperseg* is determined by solving
        # the following equation:
        #  n_segments * nperseg - (n_segments-1) * overlap * nperseg =
        #      data.shape[-1]
        #  -------------------    ===============================  ^^^^^^^^^^^
        # summed segment lengths        total overlap              data length
        nperseg = int(xdata.shape[axis] / (n_segments - overlap * (
            n_segments - 1)))
    params['nperseg'] = nperseg
    params['noverlap'] = int(nperseg * overlap)

    freqs, Pxx = scipy.signal.welch(xdata, **params)
    _, Pyy = scipy.signal.welch(ydata, **params)
    _, Pxy = scipy.signal.csd(xdata, ydata, **params)

    coherency = np.abs(Pxy) ** 2 / (Pxx * Pyy)
    phase_lag = np.angle(Pxy)

    # attach proper units to return values
    if isinstance(signal_i, pq.quantity.Quantity):
        freqs = freqs * pq.Hz
        phase_lag = phase_lag * pq.rad

    # When the input is AnalogSignal, the axis for frequency index is
    # rolled to the first to comply with the Neo convention about time axis
    if isinstance(signal_i, neo.AnalogSignal):
        coherency = np.rollaxis(coherency, -1)
        phase_lag = np.rollaxis(phase_lag, -1)

    return freqs, coherency, phase_lag


def welch_cohere(*args, **kwargs):
    warnings.warn("'welch_cohere' is deprecated; use 'welch_coherence'",
                  DeprecationWarning)
    return welch_coherence(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Unit tests for the causality module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
from __future__ import division, print_function

import unittest

import numpy as np
import quantities as pq
from neo.core import AnalogSignal
from numpy.testing import assert_array_almost_equal

from elephant.spectral import multitaper_cross_spectrum, multitaper_coherence
import elephant.causality.granger
from elephant.datasets import download_datasets


class PairwiseGrangerTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        np.random.seed(1)
        cls.ground_truth = cls._generate_ground_truth()

    @staticmethod
    def _generate_ground_truth(length_2d=30000):
        order = 2
        signal = np.zeros((2, length_2d + order))

        weights_1 = np.array([[0.9, 0], [0.9, -0.8]])
        weights_2 = np.array([[-0.5, 0], [-0.2, -0.5]])

        weights = np.stack((weights_1, weights_2))

        noise_covariance = np.array([[1., 0.0], [0.0, 1.]])

        for i in range(length_2d):
            for lag in range(order):
                signal[:, i + order] += np.dot(weights[lag],
                                               signal[:, i + 1 - lag])
            rnd_var = np.random.multivariate_normal([0, 0],
                                                    noise_covariance)
            signal[:, i+order] += rnd_var

        signal = signal[:, 2:]

        # Return signals as Nx2
        return signal.T

    def setUp(self):
        # Generate a smaller random dataset for tests other than ground truth,
        # using a different seed than in the ground truth - the convergence
        # should not depend on the seed.
        np.random.seed(10)
        self.signal = self._generate_ground_truth(length_2d=1000)

        # Estimate Granger causality
        self.causality = elephant.causality.granger.pairwise_granger(
            self.signal, max_order=10,
            information_criterion='bic')

    def test_analog_signal_input(self):
        """
        Check if analog signal input result matches an otherwise identical 2D
        numpy array input result.
        """
        analog_signal = AnalogSignal(self.signal, units='V',
                                     sampling_rate=1*pq.Hz)
        analog_signal_causality = \
            elephant.causality.granger.pairwise_granger(
                analog_signal, max_order=10,
                information_criterion='bic')
        self.assertEqual(analog_signal_causality.directional_causality_x_y,
                         self.causality.directional_causality_x_y)
        self.assertEqual(analog_signal_causality.directional_causality_y_x,
                         self.causality.directional_causality_y_x)
        self.assertEqual(analog_signal_causality.instantaneous_causality,
                         self.causality.instantaneous_causality)
        self.assertEqual(analog_signal_causality.total_interdependence,
                         self.causality.total_interdependence)

    def test_aic(self):
        identity_matrix = np.eye(2, 2)
        self.assertEqual(elephant.causality.granger._aic(
            identity_matrix, order=2, dimension=2, length=2
        ), 8.0)

    def test_bic(self):
        identity_matrix = np.eye(2, 2)
        assert_array_almost_equal(elephant.causality.granger._bic(
            identity_matrix, order=2, dimension=2, length=2
        ), 5.54517744, decimal=8)

    def test_lag_covariances_error(self):
        """
        Check that if a signal length is shorter than the set max_lag, a
        ValueError is raised.
        """
        short_signals = np.array([[1, 2], [3, 4]])
        self.assertRaises(ValueError,
                          elephant.causality.granger._lag_covariances,
                          short_signals, dimension=2, max_lag=3)

    def test_pairwise_granger_error_null_signals(self):
        null_signals = np.array([[0, 0], [0, 0]])
        self.assertRaises(ValueError,
                          elephant.causality.granger.pairwise_granger,
                          null_signals, max_order=2)

    def test_pairwise_granger_identical_signal(self):
        same_signal = np.hstack([self.signal[:, 0, np.newaxis],
                                 self.signal[:, 0, np.newaxis]])
        self.assertRaises(ValueError,
                          elephant.causality.granger.pairwise_granger,
                          signals=same_signal, max_order=2)

    def test_pairwise_granger_error_1d_array(self):
        array_1d = np.ones(10, dtype=np.float32)
        self.assertRaises(ValueError,
                          elephant.causality.granger.pairwise_granger,
                          array_1d, max_order=2)

    def test_result_namedtuple(self):
        """
        Check if the result of pairwise_granger is in the form of namedtuple.
        """
        # Import the namedtuple class for the result formatting
        from elephant.causality.granger import Causality

        # Check that the output matches the class
        self.assertIsInstance(self.causality, Causality)

    def test_result_directional_causalities_not_negative(self):
        """
        The directional causalities should never be negative.
        """
        self.assertTrue(self.causality.directional_causality_x_y >= 0)
        self.assertTrue(self.causality.directional_causality_y_x >= 0)

    def test_result_instantaneous_causality_not_negative(self):
        """
        The time-series granger instantaneous causality should never assume
        negative values.
        """
        self.assertTrue(self.causality.instantaneous_causality >= 0)

    def test_total_channel_interdependence_equals_sum_of_other_three(self):
        """
        Test if total interdependence is equal to the sum of the other three
        measures. It should be equal. In this test, however, almost equality
        is asserted due to a loss of significance with larger datasets.
        """
        causality_sum = self.causality.directional_causality_x_y \
            + self.causality.directional_causality_y_x \
            + self.causality.instantaneous_causality
        assert_array_almost_equal(self.causality.total_interdependence,
                                  causality_sum, decimal=2)

    def test_all_four_result_values_are_floats(self):
        self.assertIsInstance(self.causality.directional_causality_x_y,
                              float)
        self.assertIsInstance(self.causality.directional_causality_y_x,
                              float)
        self.assertIsInstance(self.causality.instantaneous_causality,
                              float)
        self.assertIsInstance(self.causality.total_interdependence, float)

    def test_ground_truth_vector_autoregressive_model(self):
        """
        Test the output of _optimal_vector_arm against the output of R vars
        generated using VAR(t(signal), lag.max=10, ic='AIC').
        """
        # First equation coefficients from R vars
        first_y1_l1 = 0.8947573989
        first_y2_l1 = -0.0003449514
        first_y1_l2 = -0.4934377020
        first_y2_l2 = -0.0018548490

        # Second equation coefficients from R vars
        second_y1_l1 = 9.009503e-01
        second_y2_l1 = -8.124731e-01
        second_y1_l2 = -1.871460e-01
        second_y2_l2 = -5.012730e-01

        coefficients, _, _ = elephant.causality.granger._optimal_vector_arm(
            self.ground_truth.T, dimension=2, max_order=10,
            information_criterion='aic')

        # Arrange the ground truth values in the same shape as coefficients
        ground_truth_coefficients = np.asarray(
            [[[first_y1_l1, first_y2_l1],
              [second_y1_l1, second_y2_l1]],
             [[first_y1_l2, first_y2_l2],
              [second_y1_l2, second_y2_l2]]]
        )

        assert_array_almost_equal(coefficients, ground_truth_coefficients,
                                  decimal=4)

    def test_wrong_kwarg_optimal_vector_arm(self):
        wrong_ic_criterion = 'cic'

        self.assertRaises(ValueError,
                          elephant.causality.granger._optimal_vector_arm,
                          self.ground_truth.T, 2, 10, wrong_ic_criterion)


class ConditionalGrangerTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        np.random.seed(1)
        cls.ground_truth = cls._generate_ground_truth()

    @staticmethod
    def _generate_ground_truth(length_2d=30000, causality_type="indirect"):
        """
        Recreated from Example 2 section 5.2 of :cite:'granger-Ding06-0608035'.
        The following should generate three signals in one of the two ways:
         1. "indirect" would generate data which contains no direct
        causal influence from Y to X, but mediated through Z
        (i.e. Y -> Z -> X).
        2. "both" would generate data which contains both direct and indirect
        causal influences from Y to X.

        """
        if causality_type == "indirect":
            y_t_lag_2 = 0
        elif causality_type == "both":
            y_t_lag_2 = 0.2
        else:
            raise ValueError("causality_type should be either 'indirect' or "
                             "'both'")

        order = 2
        signal = np.zeros((3, length_2d + order))

        weights_1 = np.array([[0.8, 0, 0.4],
                              [0, 0.9, 0],
                              [0., 0.5, 0.5]])

        weights_2 = np.array([[-0.5, y_t_lag_2, 0.],
                              [0., -0.8, 0],
                              [0, 0, -0.2]])

        weights = np.stack((weights_1, weights_2))

        noise_covariance = np.array([[0.3, 0.0, 0.0],
                                     [0.0, 1., 0.0],
                                     [0.0, 0.0, 0.2]])

        for i in range(length_2d):
            for lag in range(order):
                signal[:, i + order] += np.dot(weights[lag],
                                               signal[:, i + 1 - lag])
            rnd_var = np.random.multivariate_normal([0, 0, 0],
                                                    noise_covariance)
            signal[:, i + order] += rnd_var

        signal = signal[:, 2:]

        # Return signals as Nx3
        return signal.T

    def setUp(self):
        # Generate a smaller random dataset for tests other than ground truth,
        # using a different seed than in the ground truth - the convergence
        # should not depend on the seed.
        np.random.seed(10)
        self.signal = self._generate_ground_truth(length_2d=1000)

        # Generate a small dataset for containing both direct and indirect
        # causality.
        self.non_zero_signal = self._generate_ground_truth(
            length_2d=1000, causality_type="both")
        # Estimate Granger causality
        self.conditional_causality = elephant.causality.granger.\
            conditional_granger(self.signal, max_order=10,
                                information_criterion='bic')

    def test_result_is_float(self):
        self.assertIsInstance(self.conditional_causality, float)

    def test_ground_truth_zero_value_conditional_causality(self):
        self.assertEqual(elephant.causality.granger.conditional_granger(
            self.ground_truth, 10, 'bic'), 0.0)

    def test_ground_truth_zero_value_conditional_causality_anasig(self):
        signals = AnalogSignal(self.ground_truth, sampling_rate=1*pq.Hz,
                               units='V')
        self.assertEqual(elephant.causality.granger.conditional_granger(
            signals, 10, 'bic'), 0.0)

    def test_non_zero_conditional_causality(self):
        self.assertGreater(elephant.causality.granger.conditional_granger(
            self.non_zero_signal, 10, 'bic'), 0.0)

    def test_conditional_causality_wrong_input_shape(self):
        signals = np.random.normal(0, 1, (4, 10, 1))

        self.assertRaises(ValueError,
                          elephant.causality.granger.conditional_granger,
                          signals, 10, 'bic')


class PairwiseSpectralGrangerTestCase(unittest.TestCase):
    def test_bracket_operator_one_signal(self):
        # Generate a spectrum from random dataset and test bracket operator
        np.random.seed(10)
        n = 10
        spectrum = np.random.normal(0, 1, n)

        # Generate causal part according to The Factorization of Matricial
        # Spectral Densities', Wilson 1972, SiAM J Appl Math, Definition 1.2
        # (ii).

        spectrum_causal = np.fft.ifft(spectrum, axis=0)
        spectrum_causal[(n + 1) // 2:] = 0
        spectrum_causal[0] /= 2

        spectrum_causal_ground_truth = np.fft.fft(spectrum_causal, axis=0)

        spectrum_causal_est = elephant.causality.granger._bracket_operator(
            spectrum=spectrum,
            num_freqs=n,
            num_signals=1)

        np.testing.assert_array_almost_equal(spectrum_causal_est,
                                             spectrum_causal_ground_truth)

    def test_bracket_operator_mult_signal(self):
        # Generate a spectrum from random dataset and test bracket operator
        np.random.seed(10)
        n = 10
        num_signals = 3
        spectrum = np.random.normal(0, 1, (n, num_signals, num_signals))

        # Generate causal part according to The Factorization of Matricial
        # Spectral Densities', Wilson 1972, SiAM J Appl Math, Definition 1.2
        # (ii).

        spectrum_causal = np.fft.ifft(spectrum, axis=0)
        spectrum_causal[(n + 1) // 2:] = 0
        spectrum_causal[0] /= 2

        spectrum_causal_ground_truth = np.fft.fft(spectrum_causal, axis=0)

        # Set element below diagonal at zero frequency to zero
        spectrum_causal_ground_truth[0, 1, 0] = 0
        spectrum_causal_ground_truth[0, 2, 0] = 0
        spectrum_causal_ground_truth[0, 2, 1] = 0

        spectrum_causal_est = elephant.causality.granger._bracket_operator(
            spectrum=spectrum,
            num_freqs=n,
            num_signals=num_signals)

        np.testing.assert_array_almost_equal(spectrum_causal_est,
                                             spectrum_causal_ground_truth)

    def test_spectral_factorization(self):
        np.random.seed(11)
        n = 100
        num_signals = 2
        signals = np.random.normal(0, 1, (num_signals, n))

        _, cross_spec = multitaper_cross_spectrum(signals,
                                                  return_onesided=True)

        cross_spec = np.transpose(cross_spec, (2, 0, 1))

        cov_matrix, transfer_function = \
            elephant.causality.granger._spectral_factorization(
                cross_spec, num_iterations=100)

        cross_spec_est = np.matmul(np.matmul(transfer_function, cov_matrix),
                                   elephant.causality.granger._dagger(
                                       transfer_function))

        np.testing.assert_array_almost_equal(cross_spec, cross_spec_est)

    def test_spectral_factorization_non_conv_exception(self):
        np.random.seed(11)
        n = 10
        num_signals = 2
        signals = np.random.normal(0, 1, (num_signals, n))

        _, cross_spec = multitaper_cross_spectrum(signals,
                                                  return_onesided=True)

        cross_spec = np.transpose(cross_spec, (2, 0, 1))

        self.assertRaises(Exception,
                          elephant.causality.granger._spectral_factorization,
                          cross_spec, num_iterations=1)

    def test_spectral_factorization_initial_cond(self):
        # Cross spectrum at zero frequency must always be symmetric
        wrong_cross_spec = np.array([[[1, 2], [-1, 1]], [[1, 1], [1, 1]]])
        self.assertRaises(ValueError,
                          elephant.causality.granger._spectral_factorization,
                          wrong_cross_spec, num_iterations=10)

    def test_dagger_2d(self):
        matrix_array = np.array([[1j, 0], [2, 3]], dtype=complex)

        true_dagger = np.array([[-1j, 2], [0, 3]], dtype=complex)

        dagger_matrix_array = elephant.causality.granger._dagger(matrix_array)

        np.testing.assert_array_equal(true_dagger, dagger_matrix_array)

    def test_total_channel_interdependence_equals_transformed_coherence(self):
        np.random.seed(11)
        n = int(2**8)
        num_signals = 2
        signals = np.random.normal(0, 1, (num_signals, n))

        freqs, coh, phase_lag = multitaper_coherence(signals[0], signals[1],
                                                     len_segment=2**7,
                                                     num_tapers=2)
        f, spectral_causality = \
            elephant.causality.granger.pairwise_spectral_granger(
                signals[0], signals[1], len_segment=2**7, num_tapers=2)

        total_interdependence = spectral_causality[3]
        # Cut last frequency due to length of segment being even and
        # multitaper_coherence using the real FFT in contrast to
        # pairwise_spectral_granger which has to use the full FFT.
        true_total_interdependence = -np.log(1 - coh)[:-1]
        np.testing.assert_allclose(total_interdependence,
                                   true_total_interdependence, atol=1e-5)

    def test_pairwise_spectral_granger_against_ground_truth(self):
        """
        Test pairwise_spectral_granger using an example from Ding et al. 2006

        Please follow the link below for more details:
        https://gin.g-node.org/NeuralEnsemble/elephant-data/src/master/unittest/causality/granger/pairwise_spectral_granger  # noqa

        """

        repo_path = \
            r"unittest/causality/granger/pairwise_spectral_granger/data"

        files_to_download = [
            ("time_series.npy", "54e0b3fbd904ccb48c75228c070a1a2a"),
            ("weights.npy", "eb1fc5590da5507293c63b25b1e3a7fc"),
            ("noise_covariance.npy", "6f80ccff2b2aa9485dc9c01d81570bf5")
        ]

        downloaded_files = {}
        for filename, checksum in files_to_download:
            downloaded_files[filename] = {
                'filename': filename,
                'path': download_datasets(repo_path=f"{repo_path}/{filename}",
                                          checksum=checksum)}

        signals = np.load(downloaded_files['time_series.npy']['path'])
        weights = np.load(downloaded_files['weights.npy']['path'])
        cov = np.load(downloaded_files['noise_covariance.npy']['path'])

        # Estimate spectral Granger Causality
        f, spectral_causality = \
            elephant.causality.granger.pairwise_spectral_granger(
                signals[0], signals[1], len_segment=2**7, num_tapers=3)

        # Calculate ground truth spectral Granger Causality
        # Formulae taken from Ding et al., Granger Causality: Basic Theory and
        # Application to Neuroscience, 2006
        fn = np.linspace(0, np.pi, len(f))
        freqs_for_theo = np.array([1, 2])[:, np.newaxis] * fn
        A_theo = (np.identity(2)[np.newaxis]
                  - weights[0] * np.exp(
                      - 1j * freqs_for_theo[0][:, np.newaxis, np.newaxis]))
        A_theo -= weights[1] * np.exp(
            - 1j * freqs_for_theo[1][:, np.newaxis, np.newaxis])

        H_theo = np.array([[A_theo[:, 1, 1], -A_theo[:, 0, 1]],
                          [-A_theo[:, 1, 0], A_theo[:, 0, 0]]])
        H_theo /= np.linalg.det(A_theo)
        H_theo = np.moveaxis(H_theo, 2, 0)

        S_theo = np.matmul(np.matmul(H_theo, cov),
                           elephant.causality.granger._dagger(H_theo))

        H_tilde_xx = (H_theo[:, 0, 0] + (cov[0, 1] /
                                         cov[0, 0] * H_theo[:, 0, 1]))
        H_tilde_yy = (H_theo[:, 1, 1] + (cov[0, 1] /
                                         cov[1, 1] * H_theo[:, 1, 0]))

        true_directional_causality_y_x = np.log(S_theo[:, 0, 0].real /
                                                (H_tilde_xx
                                                 * cov[0, 0]
                                                 * H_tilde_xx.conj()).real)

        true_directional_causality_x_y = np.log(S_theo[:, 1, 1].real /
                                                (H_tilde_yy * cov[1, 1] *
                                                 H_tilde_yy.conj()).real)

        true_instantaneous_causality = np.log(
            (H_tilde_xx * cov[0, 0] * H_tilde_xx.conj()).real
            * (H_tilde_yy * cov[1, 1] * H_tilde_yy.conj()).real)
        true_instantaneous_causality -= np.linalg.slogdet(S_theo)[1]

        np.testing.assert_allclose(
            spectral_causality.directional_causality_x_y,
            true_directional_causality_x_y, atol=0.06)

        np.testing.assert_allclose(
            spectral_causality.directional_causality_y_x,
            true_directional_causality_y_x, atol=0.06)

        np.testing.assert_allclose(
            spectral_causality.instantaneous_causality,
            true_instantaneous_causality, atol=0.06)

    def test_pairwise_spectral_granger_against_r_grangers(self):
        """
        Test pairwise_spectral_granger against R grangers implementation

        Please follow the link below for more details:
        https://gin.g-node.org/NeuralEnsemble/elephant-data/src/master/unittest/causality/granger/pairwise_spectral_granger  # noqa

        """

        repo_path = \
            r"unittest/causality/granger/pairwise_spectral_granger/data"

        files_to_download = [
            ("time_series_small.npy", "b33dc12d4291db7c2087dd8429f15ab4"),
            ("gc_matrix.npy", "c57262145e74a178588ff0a1004879e2")
        ]

        downloaded_files = {}
        for filename, checksum in files_to_download:
            downloaded_files[filename] = {
                'filename': filename,
                'path': download_datasets(repo_path=f"{repo_path}/{filename}",
                                          checksum=checksum)}

        signal = np.load(downloaded_files['time_series_small.npy']['path'])
        gc_matrix = np.load(downloaded_files['gc_matrix.npy']['path'])

        denom = 20
        f, spectral_causality = \
            elephant.causality.granger.pairwise_spectral_granger(
                signal[0], signal[1], len_segment=int(len(signal[0]) / denom),
                num_tapers=15, fs=1, num_iterations=50)

        np.testing.assert_allclose(gc_matrix[::denom, 0], f, atol=4e-5)
        np.testing.assert_allclose(gc_matrix[::denom, 1],
                                   spectral_causality[0], atol=0.085)
        np.testing.assert_allclose(gc_matrix[::denom, 2],
                                   spectral_causality[1], atol=0.035)


if __name__ == '__main__':
    unittest.main()
